#! /usr/bin/env python3
# 
# Desktop search support using large language models (LLMs).
#
# note:
# - Initially based on following sources:
#   https://swharden.com/blog/2023-07-29-ai-chat-locally-with-python
#   https://github.com/alejandro-ao/ask-multiple-pdfs
#
# TODO2:
# - Make sure the tests are more reliabile (e.g., perhaps issue due to caching).
# - Clarify the specific Python requirements (language and modules)
#
# TODO3:
# - Document the workflow better.
# - Note that --index arg changed from binary to text with index dir.
#

"""
Desktop search utility

Sample usage:
    {script} --index ..

    {script} --similar "license"
"""

# Standard modules
import time
import pathlib
import atexit
from collections.abc import Iterable

# Installed modules
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.prompts import PromptTemplate
from langchain.chains import RetrievalQA
from langchain_core.documents import Document
from langchain_community.document_loaders import DirectoryLoader, TextLoader
## OLD: from langchain_community.embeddings import HuggingFaceEmbeddings
from langchain_huggingface import HuggingFaceEmbeddings
from langchain_community.llms import CTransformers
from langchain_community.vectorstores import FAISS

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import gpu_utils
from mezcla.main import Main, KEEP_TEMP_FILES
from mezcla import system
from mezcla import html_utils
from mezcla import extract_document_text
from mezcla.my_regex import my_re

# Constants
TL = debug.TL
INDEX_ARG = "index"
SEARCH_ARG = "search"
SIMILAR_ARG = "similar"
TORCH_DEVICE = system.getenv_text(
    "TORCH_DEVICE", gpu_utils.TORCH_DEVICE,
    desc="Torch devcice to use--cuda or mps if available else cpu")
debug.trace_expr(5, TORCH_DEVICE)
QA_TEMPLATE = """Use the following pieces of information to answer the user's question.
If you don't know the answer, just say that you don't know, don't try to make up an answer.
Context: {context}
Question: {question}
Only return the helpful answer below and nothing else.
Helpful answer:
"""

# Environment options
NUM_SIMILAR = system.getenv_int(
    "NUM_SIMILAR", 10,
    description="Number of similar documents to show")
HOME_DIR = gh.HOME_DIR
EMBEDDING_MODEL = system.getenv_text(
    "EMBEDDING_MODEL", "sentence-transformers/all-MiniLM-L6-v2",
    description="Model for sentence transformer embeddings")
# note: changed default to currrent dir rather than home
QA_LLM_DEFAULT = "llama-2-7b-chat.ggmlv3.q8_0.bin"
QA_LLM_MODEL = system.getenv_text(
    "QA_LLM_MODEL", QA_LLM_DEFAULT,
    description="Path to LLM model file for Q&A, such as a llama-2-7b .bin file")
QA_LLM_TYPE = system.getenv_text(
    "QA_LLM_TYPE", "llama",
    description="Type of transformer model for Q&A, such as llama or gpt2")
ALLOW_UNSAFE_MODELS_DEFAULT = True
ALLOW_UNSAFE_MODELS = system.getenv_value(
    "ALLOW_UNSAFE_MODELS", ALLOW_UNSAFE_MODELS_DEFAULT,
    description="Whether to allow loading of possibly unsafe Pickle models")
CHUNK_SIZE = system.getenv_int(
    "CHUNK_SIZE", 500,
    description="Number of characters for text splitter chunking")
CHUNK_OVERLAP = system.getenv_int(
    "CHUNK_OVERLAP", 50,
    description="Size of overlap in characters for text splitter chunking")
MAX_NEW_TOKENS = system.getenv_int(
    "MAX_NEW_TOKENS", 256,
    description="Max number of generated tokens for Q&A LLM")
TEMPERATURE = system.getenv_int(
    "TEMPERATURE", 0.01,
    description="Degree of randomness or creativity of generated text")
CONTEXT_LENGTH = system.getenv_int(
    "CONTEXT_LENGTH", 512,
    description="Context window size in tokens")
GPU_LAYERS = system.getenv_int(
    "GPU_LAYERS", (0 if (TORCH_DEVICE == "cpu") else -1),
    description="Number of layers to use for CTransfomers model")
INDEX_STORE_DIR = system.getenv_text(
    # note: changed default to current directory to avoid script dir being reused
    "INDEX_STORE_DIR", "faiss",
    description="path to store index data base")
INDEX_ONLY_RECENT = system.getenv_bool(
    "INDEX_ONLY_RECENT", True,
    description="whether or not to filter files by modification time newer than index")


def get_file_mod_fime(path: str) -> float:
    """Returns file modification time in fractional seconds
    Note: -1 is returned if the file doesn't exist
    """
    # EX: (get_file_mod_fime("/etc/passwd") > get_file_mod_fime("/boot/vmlinuz"))
    # EX: get_file_mod_fime(" /etc /password") => -1
    result = (system.get_file_modification_time(path, as_float=True) or -1)
    debug.trace(7, f"get_file_mod_fime({path}) => {result}")
    return result


def get_last_modified_date(iterable: Iterable) -> float:
    """return the newest modification date as a float, 
       or -1 if iterable is empty or files don't exist"""
    result = -1
    if iterable:
        result = max(map(get_file_mod_fime, iterable))
    return result

def correct_metadata(doc: Document, base_dir: str) -> Document:
    """removes BASE_DIR from the source metadata path
       so it represents the actual source of the file"""
    # TODO3: fix docstrings with respect to hardcoded assumptions
    debug.trace_expr(4, base_dir)
    doc_metadata = doc.metadata
    old_source = doc_metadata['source']
    debug.trace_expr(4, old_source)
    new_source = old_source.replace(base_dir, "")
    debug.assertion(new_source != old_source)
    debug.trace_expr(4, new_source)
    doc_metadata['source'] = new_source
    doc.metadata = doc_metadata
    debug.trace(5, f"correct_metadata({doc!r}, {base_dir!r}) => {doc!r}")
    return doc

def convert_to_txt(in_file: str) -> str:
    """reads non-txt files and returns the text inside them"""
    text = ""
    try:
        if in_file.endswith('.html'):
            text = html_utils.html_to_text(system.read_file(in_file))
        else:
            text = extract_document_text.document_to_text(in_file)
    except:
        debug.trace_exception(6, "convert_to_txt")
    debug.trace(7, f"convert_to_txt({in_file}) => {text!r}")
    return text


class DesktopSearch:
    """Class for searching local computer"""

    def __init__(self, index_store_dir=None):
        """Initializer; index placed in INDEX_STORE_DIR"""
        debug.trace_fmtd(TL.VERBOSE, "Helper.__init__(): self={s}", s=self)
        if index_store_dir is None:
            index_store_dir = INDEX_STORE_DIR
        self.index_store_dir = index_store_dir
        self.embeddings = None
        self.db = None
        self.llm = None
        self.qa_llm = None
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")

    def index_dir(self, dir_path):
        """Index files at DIR_PATH"""
        ## TODO4: look into indexing files from buffers rather than external files
        debug.trace(4, f"DesktopSearch.index_dir({dir_path})")

        # Make sure target index directory exists
        if not system.is_directory(self.index_store_dir):
            gh.full_mkdir(self.index_store_dir)
        
        # define what documents to load
        text_loader_kwargs={'autodetect_encoding': True}
        loader = DirectoryLoader(dir_path, glob="./*.txt", loader_cls=TextLoader, loader_kwargs=text_loader_kwargs)
        debug.trace_expr(5, loader)
        
        # copy files over to temp dir
        # note: timestamp strips the time of day (e.g., 2024-05-11)
        timestamp = debug.timestamp().split(' ', maxsplit=1)[0]
        real_path = system.real_path(dir_path)
        # note: using [1:] to remove the initial path separator 
        temp_base = system.form_path(system.TEMP_DIR, f"llm_desktop_search.{timestamp}")
        temp_path = system.form_path(temp_base,real_path[1:])
        gh.full_mkdir(temp_path)
        
        list_files = system.get_directory_filenames(real_path)
        filtered_files = list_files
        # filter files by modification time if needed
        # note: The modification time is -1 
        modif_time = get_last_modified_date(system.get_directory_filenames(self.index_store_dir))
        if INDEX_ONLY_RECENT:
            filtered_files = [f for f in list_files if (get_file_mod_fime(f) > modif_time)]
        
        files_to_convert = [found for found in filtered_files if my_re.match(r'.*\.(pdf|docx|html|txt)', found)]
        # register cleanup function before creating temp files
        if not KEEP_TEMP_FILES:
            atexit.register(gh.delete_directory, temp_path)
        for num, file in enumerate(files_to_convert):
            filename = system.filename_proper(file)
            file_tmp_path = system.form_path(temp_path, filename) 
            if file.endswith('.txt'):
                system.write_file(file_tmp_path, system.read_entire_file(file, encoding="unicode_escape"))
            else:
                system.write_file(f"{file_tmp_path}_temp_{num}.txt", convert_to_txt(file))
        
        # interpret information in the documents
        loader = DirectoryLoader(temp_path, glob="*.txt", loader_cls=TextLoader)
        documents = loader.load()
        debug.trace_expr(5, len(documents), documents, max_len=1024)
        splitter = RecursiveCharacterTextSplitter(chunk_size=CHUNK_SIZE,
                                                  chunk_overlap=CHUNK_OVERLAP)
        # documents are splitted to a maximum of 500 characters per chunk (by default)
        texts = splitter.split_documents(documents)
        corrected_texts = [correct_metadata(text, temp_base) for text in texts]
        if not self.embeddings:
            self.embeddings = HuggingFaceEmbeddings(
                model_name=EMBEDDING_MODEL,
                model_kwargs={'device': TORCH_DEVICE})

        # add (or create from docs) to the db and save it
        try:
            self.load_index()
        except RuntimeError:
            debug.trace_exception(6, "load_index")
        if self.db is not None:
            self.db.add_documents(corrected_texts)
        else:
            self.db = FAISS.from_documents(corrected_texts, self.embeddings)
        self.db.save_local(self.index_store_dir)

        debug.trace_expr(4, self.db)
        gpu_utils.trace_gpu_usage()

        # show index info
        num_chunks = len(self.db.index_to_docstore_id)
        print(f"{num_chunks} chunks indexed")

    def load_index(self, for_qa=False):
        """Load index of documents"""
        debug.trace(4, "DesktopSearch.load_index()")
        # load the language model
        config = {'max_new_tokens': MAX_NEW_TOKENS, 'temperature': TEMPERATURE,
                  'context_length': CONTEXT_LENGTH}
        if for_qa:
            llm = CTransformers(model=QA_LLM_MODEL, model_type=QA_LLM_TYPE,
                                config=config, gpu_layers=GPU_LAYERS)
            debug.trace_expr(4, llm)
            debug.trace_object(5, llm)
            self.llm = llm

        # load the interpreted information from the local database
        if not self.embeddings:
            self.embeddings = HuggingFaceEmbeddings(
                model_name=EMBEDDING_MODEL,
                model_kwargs={'device': TORCH_DEVICE})
        options = {}
        if ALLOW_UNSAFE_MODELS:
            options["allow_dangerous_deserialization"] = ALLOW_UNSAFE_MODELS
        self.db = FAISS.load_local(self.index_store_dir, self.embeddings, **options)
        debug.trace_expr(5, self.llm, self.embeddings, self.db)
        gpu_utils.trace_gpu_usage()

    def prepare_qa_llm(self):
        """Prepare a version of the llm pre-loaded with the local content"""
        debug.trace(4, "prepare_qa_llm()")
        if not self.db:
            self.load_index(for_qa=True)        
        retriever = self.db.as_retriever(search_kwargs={'k': NUM_SIMILAR})
        prompt = PromptTemplate(
            template=QA_TEMPLATE,
            input_variables=['context', 'question'])

        self.qa_llm = RetrievalQA.from_chain_type(
            llm=self.llm, chain_type='stuff', retriever=retriever,
            return_source_documents=True, chain_type_kwargs={'prompt': prompt})
        gpu_utils.trace_gpu_usage()
        
    def search_to_answer(self, question):         # TODO2: rename like answer_question
        """Search documents to answer QUESTION"""
        debug.trace(4, f"DesktopSearch.search_to_answer({question})")
        if not self.qa_llm:
            self.prepare_qa_llm()
        model_path = self.qa_llm.combine_documents_chain.llm_chain.llm.model
        model_name = pathlib.Path(model_path).name
        time_start = time.time()
        output = self.qa_llm({'query': question})
        response = output["result"]
        time_elapsed = time.time() - time_start
        print(f'<code>{model_name} response time: {time_elapsed:.02f} sec</code>')
        print(f'<strong>Question:</strong> {question}')
        print(f'<strong>Answer:</strong> {response}')
        gpu_utils.trace_gpu_usage()

    def show_similar(self, query, num=None):
        """Show similar documents to QUERY in the vector store, up to NUM"""
        debug.trace(4, f"DesktopSearch.show_similar(query={query}, n={num})")
        if num is None:
            num = NUM_SIMILAR
        if not self.db:
            self.load_index()
        docs = self.db.similarity_search_with_score(query=query, k=num)
        print("Similar documents:")
        for doc in docs:
            print(doc)
        gpu_utils.trace_gpu_usage()

class Script(Main):
    """Adhoc script class (e.g., no I/O loop, just run calls)"""
    # note: class-level variables for arguments avoids need for class constructor
    index_arg = None
    search_arg = False
    similar_arg = False

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(TL.VERBOSE, "Script.setup(): self={s}", s=self)
        self.index_arg = self.get_parsed_option(INDEX_ARG, self.index_arg)
        self.search_arg = self.get_parsed_option(SEARCH_ARG, self.search_arg)
        self.similar_arg = self.get_parsed_option(SIMILAR_ARG, self.similar_arg)
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")

    def run_main_step(self):
        """Main processing step (n.b., assumes self.manual_input)"""
        debug.trace_fmtd(5, "Script.run_main_step(): self={s}", s=self)
        ds = DesktopSearch()
        if self.index_arg:
            ds.index_dir(self.index_arg)
        elif self.search_arg:
            ds.search_to_answer(self.search_arg)
        elif self.similar_arg:
            ds.show_similar(self.similar_arg)
        else:
            system.print_error("Error: Unexpected condition")


def main():
    """Entry point"""
    app = Script(
        description=__doc__.format(script=gh.basename(__file__)),
        skip_input=True, manual_input=True,
        boolean_options=[],
        text_options=[(INDEX_ARG, "Index directory"),
                      (SEARCH_ARG, "Search documents to answer question"),
                      (SIMILAR_ARG, "Show similar documents")],
        float_options=None)
    app.run()

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    debug.trace(5, f"module __doc__: {__doc__}")
    main()



#! /usr/bin/env python3
#
# Use scikit-learn to perform probabilistic grid search for number of hidden
# layers and number of units per layer.
#
# This extends the following brute-force approach to be both probabilistic
# and to handle hidden layer optimization:
#    https://machinelearningmastery.com/grid-search-hyperparameters-deep-learning-models-python-keras/
# It is also based on the multiclass support from the following:
#    https://machinelearningmastery.com/multi-class-classification-tutorial-keras-deep-learning-library
#
# TODO:
# - Parameterize the following options:
#   activation, num_layers, num_units, ...
# - Fix random seed initialization.
# - Work out the stupid keras/tensorflow incompatibilies issues.
#

"""Keras for probabilistic grid search via scikit-learn"""

## TEST: try to set random number seed for TensorFlow (to get deterministic results)
## import numpy
## numpy.random.seed(7919)
## import tensorflow
## BAD: tensorflow.set_random_seed(7919)
## tensorflow.random.set_seed(7919)

# Standard moduless
import re
import sys
from collections import OrderedDict

# Installed moduless
# pylint: disable=import-error
# TODO: install kera dynamically???
## TEST: sys.stderr.write("here\n")
## BAD:
from keras.models import Sequential
## See https://stackoverflow.com/questions/55496289/how-to-fix-attributeerror-module-tensorflow-has-no-attribute-get-default-gr
## TEST: from tensorflow.keras.models import Sequential

## OLD:
##
from keras.layers import Dense
##
from keras.wrappers.scikit_learn import KerasClassifier
##
## TEST:
## from tensorflow.keras.layers import Dense
## from tensorflow.keras.wrappers.scikit_learn import KerasClassifier
##
from keras.utils import np_utils

## TEST: sys.stderr.write("there\n")
## TODO: import numpy
import pandas
from pandas.core.frame import DataFrame
from sklearn.model_selection import cross_val_score, GridSearchCV, KFold, RandomizedSearchCV
from sklearn.preprocessing import LabelEncoder

## NOTE: This takes too long to load so postponed, in case usage just shown.
## OLD: import tensorflow as tf
## TEST: tf = None

# Local moduless
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system
round3 = system.round3
from mezcla import text_utils
from mezcla.text_utils import getenv_ints

# Note: python 3.6+ format strings are used
assert((sys.version_info.major >= 3) and (sys.version_info.minor >= 6))

#...............................................................................
# Constants (e.g., based on environment)

DEFAULT_VERBOSITY = ((debug.get_level() + 1) // 2)
VERBOSITY_LEVEL = system.getenv_int("VERBOSITY_LEVEL", DEFAULT_VERBOSITY)
DATA_FILE = system.getenv_text("DATA_FILE", "examples/pima-indians-diabetes.csv")
FIELD_SEP = system.getenv_text("FIELD_SEP", ",")
BRUTE_FORCE = system.getenv_bool("BRUTE_FORCE", False)
RANDOM_OPTIMIZATION = (not BRUTE_FORCE)
SEED = system.getenv_int("SEED", 7919)
OUTPUT_CSV = system.getenv_bool("OUTPUT_CSV", False)

QUICK_SEARCH = system.getenv_bool("QUICK_SEARCH", False)
FULLER_SEARCH = (RANDOM_OPTIMIZATION and not QUICK_SEARCH)

# TODO: Add descriptions for important deep learning parameters (e.g., NUM_EPOCHS and BATCH_SIZE).
# Note: NUM_EPOCHS and BATCH_SIZE not used in grid search (just sample classification)
RUN_SAMPLE =  system.getenv_bool("RUN_SAMPLE", False)
DEFAULT_NUM_EPOCHS = (200 if (not QUICK_SEARCH) else 10)
NUM_EPOCHS = system.getenv_int("NUM_EPOCHS", DEFAULT_NUM_EPOCHS)
BATCH_SIZE = system.getenv_int("BATCH_SIZE", 5)
NUM_FOLDS = system.getenv_int("NUM_FOLDS", 10)
## OLD: NUM_JOBS = system.getenv_int("NUM_JOBS", -1, "Number of parallel jobs (-1 uses all cores)")
NUM_JOBS = system.getenv_int("NUM_JOBS", -1, "Number of parallel jobs: -1 uses all cores")
NUM_ITERS = system.getenv_int("NUM_ITERS", 100)
SCORING_METRIC = system.getenv_text("SCORING_METRIC", "accuracy")
## OLD: USE_ONE_HOT = system.getenv_bool("USE_ONE_HOT", False)
SKIP_ONE_HOT = system.getenv_bool("SKIP_ONE_HOT", False)
USE_ONE_HOT = system.getenv_bool("USE_ONE_HOT", (not SKIP_ONE_HOT))

# note: HIDDEN_UNIT_VALUES is just for grid search
DEFAULT_HIDDEN_UNIT_VALUES = ("0 5 10 25 50 75 100 250 500" if FULLER_SEARCH else "0 10 50 100 250")
HIDDEN_UNIT_VALUES = getenv_ints("HIDDEN_UNIT_VALUES", DEFAULT_HIDDEN_UNIT_VALUES)
DEFAULT_BATCH_SIZE_VALUES = ("5 25 75 100" if FULLER_SEARCH else "5 50")
BATCH_SIZE_VALUES = getenv_ints("BATCH_SIZE_VALUES", DEFAULT_BATCH_SIZE_VALUES)
DEFAULT_NUM_EPOCH_VALUES = ("10 50 100 250 500 1000" if FULLER_SEARCH else "50 250")
NUM_EPOCH_VALUES = getenv_ints("NUM_EPOCH_VALUES", DEFAULT_NUM_EPOCH_VALUES)

MAX_HIDDEN_UNIT_VARS = system.getenv_int("MAX_HIDDEN_UNIT_VARS", 5)
HIDDEN_UNIT_VARS = ["hidden_units{n}".format(n=(v + 1)) for v in range(MAX_HIDDEN_UNIT_VARS)]
# note: DEFAULT_HIDDEN_UNITS is for use outside of grid search
DEFAULT_HIDDEN_UNITS = getenv_ints("HIDDEN_UNITS", "20 30")

#...............................................................................
# Utility functions

def non_negative(num):
    """Whether integer NUM > -1"""
    return (num > -1)

# TODO: put following in new ml_utils.py module
def create_feature_mapping(label_values):
    """Return hash mapping elements from LABEL_VALUES into integers"""
    # EX: create_feature_mapping(['c', 'b', 'b', 'a']) => {'c':0, 'b':1, 'a':2}
    debug.assertion(isinstance(label_values, list))
    id_hash = {}
    for item in label_values:
        if (item not in id_hash):
            id_hash[item] = len(id_hash)
    debug.trace_fmt(7, "create_feature_mapping({lv}) => {ih}", lv=label_values, ih=id_hash)
    return id_hash

#...............................................................................
# Grid search support

def create_keras_model(num_input_features=None, num_classes=None, hidden_units=None, **kwargs):
    """Create n-layer Keras model, using either the HIDDEN_UNITS vector or via the hidden_unitsN entries of KWARGS"""
    # TODO: put under MyKerasClassifier as __call__
    debug.trace_fmt(5, "create_keras_model(#f={nf}, #c={nc}, hu={hu}, kw={kw})",
                    nf=num_input_features, nc=num_classes, hu=hidden_units, kw=kwargs)

    # Initialize defaults
    if (num_input_features is None):
        debug.trace(2, "Warning: number of features not specified so using 100!")
        num_input_features = 100
    if (num_classes is None):
        debug.trace(2, "Warning: number of classes not specified so using 2!")
        num_classes = 2
    is_binary = (num_classes == 2)
    if (hidden_units is None):
        # note: -1 needs to be specified in the estimator constructor in order for the grid search
        # variable to be recognized, therefore only non-negative counts are considered.
        hidden_unit_counts = [(kwargs.get(v) or -1) for v in HIDDEN_UNIT_VARS]
        ## OLD: num_non_negative = sum([int(non_negative(n)) for n in hidden_unit_counts])
        num_non_negative = sum(int(non_negative(n)) for n in hidden_unit_counts)
        debug.trace_fmt(5, "hidden_unit_counts={huc} num_non_negative={nnn}", huc=hidden_unit_counts, nnn=num_non_negative)
        if (num_non_negative > 0):
            hidden_units = hidden_unit_counts
            debug.trace_fmt(4, "Using hidden unit counts from kwarg vars: {hu}", hu=hidden_units)
    if (hidden_units is None):
        hidden_units = DEFAULT_HIDDEN_UNITS
        debug.trace_fmt(2, "Warning: neither HIDDEN_UNITS nor any hidden_unitsN specified so using default: {hu}", hu=hidden_units)

    # Create the model with optional hidden layers
    # TODO: parameterize activation fn
    model = Sequential()
    for (i, hidden_unit_count) in enumerate(hidden_units):
        num_inputs = num_input_features if (i == 0) else None
        model.add(Dense(hidden_unit_count, input_dim=num_inputs, activation="relu"))
    # Add output layer
    if is_binary:
        model.add(Dense(1, activation="sigmoid"))
    else:
        model.add(Dense(num_classes, activation="softmax"))

    # Compile model using Adaptive Moment Estimation (Adam) optimizer and cross-entropy loss function.
    # note: one-hot encoding apparently needed for categorical_crossentropy
    debug.assertion(is_binary or USE_ONE_HOT)
    loss_function = "binary_crossentropy" if is_binary else "categorical_crossentropy"
    model.compile(loss=loss_function, optimizer="adam", metrics=[SCORING_METRIC])
    debug.trace_object(5, model, "model")
    debug.trace_fmt(4, "create_keras_model() => {m}", m=model)

    return model


class MyKerasClassifier(KerasClassifier):
    """Defines a version of KerasClassifier that is not so picky about parameters in order 
    to support a dynamic number of hidden unit variables (i.e., without having to explicitly
    enumerate them as in __init_(self, ..., hidden_units1=-1, ... hidden_units10=-1)."""
    class_name = "MyKerasClassifier"     # TODO: derive via introspection

    def __init__(self, **kwargs):
        """Class constructor: record list off keyword arguments"""
        debug.trace_fmt(5, "{cl}.__init__(kw={kw})", cl=self.class_name, kw=kwargs)
        self.params = list(kwargs.keys())
        ## OLD: super(MyKerasClassifier, self).__init__(**kwargs)
        super().__init__(**kwargs)

    def check_params(self, params):
        """Make sure all params supported"""
        ok = (not system.difference(list(params.keys()), self.params))
        debug.trace_fmt(6, "{cl}.check_params({p}) => {r}", cl=self.class_name, p=params, r=ok)
        return ok

#................................................................................

def main():
    """Main entry point for script"""
    # TODO: convert into using main.py's Main class
    data_file = None
    args = system.get_args()
    if ((len(args) > 1) and (args[1] == "--help")):
        script = gh.basename(args[0])
        system.print_stderr("Usage: {scr} [--help] [data-file | -]", scr=script)
        system.print_stderr("")
        system.print_stderr("where data-file is in CSV format and - is for default")
        system.print_stderr("")
        ## TODO: show_details = VERBOSE or debug.debugging()
        if debug.debugging(2):
            ## TODO: add legend
            system.print_stderr("\t" + system.formatted_environment_option_descriptions(include_all=True))
        system.exit("")
    if ((len(args) > 1) and (not args[1].startswith("-"))):
        data_file = args[1]
    debug.trace(4, f"data_file={data_file}")
        
    # load dataset
    # TODO: only skip first row if all symbolic
    ## OLD: headers = system.read_entire_file(DATA_FILE).split("\n")[0].split(FIELD_SEP)
    if data_file is None:
        data_file = gh.resolve_path(DATA_FILE)
    headers = system.read_entire_file(data_file).split("\n")[0].split(FIELD_SEP)
    debug.assertion(all(text_utils.is_symbolic(v) for v in headers))
    ## OLD: dataset = numpy.loadtxt(DATA_FILE, delimiter=FIELD_SEP, skiprows=1)
    ## OLD: data_frame = pandas.read_csv(DATA_FILE, sep=FIELD_SEP)
    data_frame = pandas.read_csv(data_file, sep=FIELD_SEP)
    dataset = data_frame.values
    
    # split into input (X) and output (y) variables
    num_features = (dataset.shape[1] - 1)
    debug.assertion(len(headers) == (num_features + 1))
    debug.assertion(num_features > 1)
    ## OLD: X = dataset[:, 0:num_features]
    X = dataset[:, 0:num_features].astype(float)
    y = dataset[:, num_features]
    if OUTPUT_CSV:
        # TODO: drop pandas index column (first one; no header)
        ## OLD: basename = system.remove_extension(DATA_FILE)
        basename = system.remove_extension(data_file)
        ## BAD:
        ## X.to_csv(basename + "-X.csv", sep=FIELD_SEP)
        ## y.to_csv(basename + "-y.csv", sep=FIELD_SEP)
        ## TODO: rework so that X and y kept as data frames; remove extraneous quotes from headers (e.g., """some field""" => "some field")
        DataFrame(X).to_csv(basename + "-X.csv", header=headers[:-1], sep=FIELD_SEP)
        DataFrame(y).to_csv(basename + "-y.csv", header=headers[-1:], sep=FIELD_SEP)
    ## TEST:
    y = list(y)
    debug.trace_fmtd(7, "X={X}\ny={y}", X=X, y=y)
    y_hash = create_feature_mapping(y)
    num_categories = len(y_hash)

    # Encode class values as integers, using one-hot vectors (i.e., one vector per category).
    symbolic_classes = all(text_utils.is_symbolic(v) for v in y)
    modified_y = y
    if symbolic_classes:
        # TODO: use inverse_transform later when analyzing the results
        encoder = LabelEncoder()
        encoder.fit(modified_y)
        modified_y = encoder.transform(y)
        debug.trace_fmtd(7, "encoded_y={ey}", ey=modified_y)
        
        # Convert integers to dummy variables (i.e,. one hot encoded)
        if USE_ONE_HOT:
            modified_y = np_utils.to_categorical(modified_y)
            debug.trace_fmtd(7, "one_hot_y={ohy}", ohy=modified_y)
        debug.trace_fmtd(8, "modified_y={my}", my=modified_y)

    # Create initial model
    dummy_hidden_unit_params = {v:-1 for v in HIDDEN_UNIT_VARS}
    # pylint: disable=unnecessary-lambda-assignment
    create_model_fn = lambda: create_keras_model(num_input_features=num_features,
                                                 num_classes=num_categories)
    ## OLD: model = KerasClassifier(build_fn=create_model_fn, verbose=VERBOSITY_LEVEL)
    ## TODO: see why batch_size needed for better accuracy
    # Note: all grid-search parameters need to be specified in the classifier constructor call.
    # Therefore, the hidden units are given -1 values.
    model = MyKerasClassifier(build_fn=create_model_fn, epochs=NUM_EPOCHS, batch_size=BATCH_SIZE,
                              verbose=VERBOSITY_LEVEL, **dummy_hidden_unit_params)
    
    # Run standard classificaion
    # note: Used for comparison against samples/keras_multiclass.py.
    ## OLD: estimator = KerasClassifier(build_fn=baseline_model, epochs=200, batch_size=5, verbose=0)
    if RUN_SAMPLE:
        try:
            kfold = KFold(n_splits=NUM_FOLDS, shuffle=True)
            results = cross_val_score(model, X, modified_y, cv=kfold)
            print("{k}-fold cross validation results:".format(k=NUM_FOLDS))
            print("Baseline: mean={m} stdev={s}; num_epochs={ne} batch_size={bs}".format(
                m=round3(results.mean()), s=round3(results.std()), ne=NUM_EPOCHS, bs=BATCH_SIZE))
        except:
            debug.trace_fmtd(2, "Error: Problem during cross_val_score: {exc}", exc=sys.exc_info())
            debug.raise_exception(6)
    else:
        debug.trace(5, "Skipped sample invocation")

    # Define the grid search parameters and then run the search with all cores and
    # 3-fold cross validation.
    debug.assertion(all(isinstance(v, list) for v in [HIDDEN_UNIT_VALUES, BATCH_SIZE_VALUES, NUM_EPOCH_VALUES]))
    hidden_unit_params = {v: HIDDEN_UNIT_VALUES for v in HIDDEN_UNIT_VARS}
    parameters = {"batch_size": BATCH_SIZE_VALUES,
                  "epochs": NUM_EPOCH_VALUES}
    parameters.update(hidden_unit_params)
    debug.trace_fmt(4, "parameters: {p}", p=parameters)

    ## OLD: grid = RandomizedSearchCV(model, parameters, n_jobs=-1, cv=3)
    # Note: much better results with 10-fold cross validation (vs. 3-fold)
    try:
        if BRUTE_FORCE:
            grid = GridSearchCV(model, parameters, n_jobs=NUM_JOBS, cv=NUM_FOLDS, error_score=0, verbose=VERBOSITY_LEVEL)
        else:
            grid = RandomizedSearchCV(model, parameters, n_jobs=NUM_JOBS, n_iter=NUM_ITERS, cv=NUM_FOLDS, error_score=0, verbose=VERBOSITY_LEVEL)
        ## OLD: grid_result = grid.fit(X, y)
        ## OLD2: grid_result = grid.fit(X, dummy_y)
        grid_result = grid.fit(X, modified_y)
        debug.trace_object(5, grid_result, "grid_result")
    except:
        debug.trace_fmtd(2, "Error: Problem during hyperparameter search: {exc}", exc=sys.exc_info())
        debug.raise_exception(6)

    # Summarize (randomized) parameter search results
    try:
        gridsearch_type = "Randomized" if (not BRUTE_FORCE) else "Brute-force"
        print("{gt} gridsearch results:".format(gt=gridsearch_type))
        ## OLD: print("Best: %f using %s" % (grid_result.best_score_, grid_result.best_params_))
        means = grid_result.cv_results_["mean_test_score"]
        stds = grid_result.cv_results_["std_test_score"]
        params = grid_result.cv_results_["params"]
        best_sorted_keys = sorted(grid_result.best_params_.keys())
        print("Best: {score} using {params}",
              score=grid_result.best_score_,
              params=OrderedDict([(k, params[k]) for k in best_sorted_keys]))
        ## OLD: print("Mean\tStdev\tParam")
        print("Metric\t\tParameters")
        header_spec = "\t".join(["Mean", "Stdev"] + best_sorted_keys)
        print(re.sub("(_units|_size)", "", header_spec))
        for mean, stdev, param in zip(means, stds, params):
            ## OLD: print("{m}\t{s}\t{p}".format(m=round3(mean), s=round3(stdev), p=param))
            param_value_spec = "\t".join([str(param[v]) for v in best_sorted_keys])
            # Hack: make sure parameter names have length < 8 (for proper tabbing)
            # TODO: define mapping with optional user override (e.g., "batch_size=>bsize, hidden_units=> hunits")
            print("{m}\t{s}\t{pvs}".format(m=round3(mean), s=round3(stdev), pvs=param_value_spec))
    except:
        debug.trace_fmtd(2, "Error: Problem during summarization: {exc}", exc=sys.exc_info())
        debug.raise_exception(6)

## TEST
## # Load tensorflow
## import tensorflow as tf                 # pylint: disable=import-outside-toplevel, import-error

## OLD:
## if debug.verbose_debugging():
##     logging_level=tf.compat.v1.logging.DEBUG
##     debug.trace_fmt(1, "setting Tensorflow debug logging (level={ll})", ll=logging_level)
##     tf.compat.v1.logging.set_verbosity(logging_level)
##     tf.compat.v1.logging.debug('tf...logging.debug test')

## TEST
# OLD: Initialize random seed for reproducibility
## numpy.random.seed(SEED)
## tf.set_random_seed(SEED)

#...............................................................................

if __name__ == "__main__":
    main()



#! /usr/bin/env python3
#
# Uses the Hugging Face API for automatic speech recognition (ASR).
#
# Based on following:
# https://stackoverflow.com/questions/71568142/how-can-i-extract-and-store-the-text-generated-from-an-automatic-speech-recognit # pylint: disable=line-too-long
#
# TODO:
# - Add chunking to handle large file:
#     https://huggingface.co/blog/asr-chunking
#

"""Speech recognition via Hugging Face

Example:

{dir}/{script} {dir}/fuzzy-testing-1-2-3.wav

USE_INTERFACE=1 {script} -
"""

# Standard modules
# TODO: import re

# Intalled modules
# Note: done dynamically below

# Local modules
from mezcla import debug
from mezcla.main import Main
from mezcla.my_regex import my_re
from mezcla import system
from mezcla import glue_helpers as gh

# Constants
TL = debug.TL

# Constants and Environment options
ASR_TASK = "automatic-speech-recognition"
# TODO: WHISPER = getenv...("whisper-large"); DEFAULT_MODEL = ...
DEFAULT_MODEL = "facebook/s2t-medium-librispeech-asr"
ASR_MODEL = system.getenv_text(
    "ASR_MODEL", DEFAULT_MODEL,
    description="Hugging Face model for ASR")
## OLD:
## USE_GPU = system.getenv_bool("USE_GPU", False, description="Uses Torch on GPU if True")
## TORCH_DEVICE_DEFAULT = ("cuda" if USE_GPU else "cpu")
## TORCH_DEVICE = system.getenv_text(
##     "TORCH_DEVICE", TORCH_DEVICE_DEFAULT,
##     description="Torch device to use")

SOUND_FILE = system.getenv_text("SOUND_FILE", "fuzzy-testing-1-2-3.wav",
                                "Audio file with speech to recognize")
USE_INTERFACE = system.getenv_bool("USE_INTERFACE", False,
                                   "Use web-based interface via gradio")

# Globals
TORCH_DEVICE = None
torch = None

#-------------------------------------------------------------------------------

# Optionally load UI support
## TODO2: rework following hugging_face_translation.py
gr = None                               # pylint: disable=invalid-name
if USE_INTERFACE:
    import gradio as gr                 # pylint: disable=import-error

def init_torch_etc():
    """Load in supporting packages like torch
    note: returns torch module object for sake of clients
    """
    # pylint: disable=redefined-outer-name, import-outside-toplevel
    debug.trace(4, "init_torch_etc()")

    # Import torch
    global torch
    import torch                        # pylint: disable=import-outside-toplevel
    HAS_CUDA = torch.cuda.is_available()
    debug.trace_expr(5, torch.__version__, HAS_CUDA)

    # Determine device to use, with fallack to CPU if no cuda
    ## TODO2: rename TORCH_DEVICE as torch_device (because not a constant)
    global TORCH_DEVICE
    USE_CPU = system.getenv_bool(
        "USE_CPU", False,
        description="Uses Torch on CPU if True")
    TORCH_DEVICE_DEFAULT = ("cpu" if (USE_CPU or not HAS_CUDA) else "cuda")
    TORCH_DEVICE = system.getenv_text(
        "TORCH_DEVICE", TORCH_DEVICE_DEFAULT,
        description="Torch device to use")
    return torch

def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    # Show simple usage if --help given
    ## TODO2: rework script_dir fixup to os.path.delim or better yet split_path
    script_dir = gh.dirname(__file__)
    if script_dir.startswith(system.real_path(".")):
        script_dir = my_re.sub(fr"{system.real_path('.')}/?", "", script_dir)
    doc = __doc__.format(script=gh.basename(__file__),
                         dir=script_dir)
    dummy_app = Main(description=doc, skip_input=False, manual_input=False)

    # Resolve path for file
    sound_file = SOUND_FILE
    if not system.file_exists(sound_file):
        script_dir = gh.dirname(__file__)
        sound_file = gh.resolve_path(SOUND_FILE, base_dir=script_dir)
    if not system.file_exists(sound_file):
        system.exit(f"Error: unable to find SOUND_FILE '{sound_file}'")
    
    # Load "heavy" packages (delayed for sake of quicker usage)
    init_torch_etc()
    from transformers import pipeline   # pylint: disable=import-outside-toplevel

    # Load model
    device = torch.device(TORCH_DEVICE)
    model = pipeline(task=ASR_TASK, model=ASR_MODEL, device=device)

    # Show UI or transcribe input file
    if USE_INTERFACE:
        pipeline_if = gr.Interface.from_pipeline(
            model,
            title="Automatic Speech Recognition (ASR)",
            description="Using pipeline with default",
            examples=[sound_file])
        pipeline_if.launch()
    else:
        print((model(sound_file))["text"])
    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# Introspection support based on alexmojaki/executing package, such as used
# in icecream expression printer.
#
# Note:
# - To avoid circular dependencies this does not use debug.py.
# - Repos with source of code adapted here:
#   https://github.com/alexmojaki/executing
#   https://github.com/gruns/icecream
# - Other repos consulted:
#   https://github.com/samuelcolvin/python-devtools
#
# TODO:
# - Place within debug.py to allow for tracing and to avoid redundant functions.
#
#................................................................................
# MIT License
# 
# Portions Copyright (c) 2019 Alex Hall, (c) 2018 Ansgar Grunseid
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Module for introspection of arguments and source code
"""

# Standard packages
import ast
import sys
from datetime import datetime
import inspect
from os.path import basename, realpath
from types import FrameType

# Local packages

# Installed packages
# note: asttokens loaded dynamically when executed (so checked here for clients)
import asttokens                        # pylint: disable=unused-import
import pprint
from textwrap import dedent
import executing


# Globals
_absent = object()
## OLD (droppped for sake of mypy):
## intro = None


def stderr_print(*args):
    """
    prints args to sys.stderr
    """
    print(*args, file=sys.stderr)


def is_literal(s):
    """
    Returns whether argument is a literal or not
    """
    try:
        ast.literal_eval(s)
    except:
        return False
    return True


## OLD: DEFAULT_LINE_WRAP_WIDTH = 70  # Characters.
DEFAULT_LINE_WRAP_WIDTH = 256            # Characters.
DEFAULT_CONTEXT_DELIMITER = "; "
DEFAULT_ARG_TO_STRING_FUNCTION = pprint.pformat
DEFAULT_OUTPUT_FUNCTION = stderr_print


def call_or_value(obj):
    """
    Returns the value of calling obj if it's callable, else returns obj as is
    """
    return obj() if callable(obj) else obj


class Source(executing.Source):
    """
    A class that extends the functionality of the `executing.Source` class to provide
    additional methods for handling and manipulating source code text with proper indentation.

    Methods
    -------
    get_text_with_indentation(node)
        Retrieves the text representation of the given AST node, preserving its indentation.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        ## DEBUG: debug.trace_expr(TL.VERBOSE, self, args, kwargs, delim="\n\t", prefix="in {Source.__init__({a})")
        ## DEBUG: debug.trace_object(5, self, label="Source instance")

    def get_text_with_indentation(self, node):
        """
        Retrieves the text representation of the given AST node, preserving its indentation.
        """
        result = self.asttokens().get_text(node)
        ## DEBUG: sys.stderr.write(f"{node=} {result=}\n")
        if "\n" in result:
            result = " " * node.first_token.start[1] + result
            result = dedent(result)
        result = result.strip()
        return result

    @classmethod
    def for_frame(cls, frame, use_cache=True):
        return super(Source, cls).for_frame(frame, use_cache)


def prefix_lines(prefix: str, s: str, start_at_line=0) -> list[str]:
    """
    Separates every word in a string,
    adds a prefix and returns them as a list
    """
    lines = s.splitlines()

    for i in range(start_at_line, len(lines)):
        lines[i] = prefix + lines[i]

    return lines


def prefix_first_line_indent_remaining(prefix: str, s: str) -> list[str]:
    """
    Adds prefix to first line and indents the remaining,
    returns the lines as a list
    """
    indent = " " * len(prefix)
    lines = prefix_lines(indent, s, start_at_line=1)
    lines[0] = prefix + lines[0]
    return lines


def format_pair(prefix: str, arg, value):
    """
    Format the pair of argument and value and add prefix, return formatted string
    """
    if arg is _absent:
        arg_lines = []
        value_prefix = prefix
    else:
        arg_lines = prefix_first_line_indent_remaining(prefix, arg)
        value_prefix = arg_lines[-1] + "="

    looks_like_a_string = value[0] + value[-1] in ["''", '""']
    if looks_like_a_string:  # Align the start of multiline strings.
        value_lines = prefix_lines(" ", value, start_at_line=1)
        value = "\n".join(value_lines)

    value_lines = prefix_first_line_indent_remaining(value_prefix, value)
    lines = arg_lines[:-1] + value_lines
    return "\n".join(lines)


def argument_to_string(obj) -> str:
    """
    Converts argument to string using `DEFAULT_ARG_TO_STRING_FUNCTION` and preserves newlines
    """
    s = DEFAULT_ARG_TO_STRING_FUNCTION(obj)
    s = s.replace("\\n", "\n")  # Preserve string newlines in output.
    return s

def trace_frame(frame, label="frame"):
    """Trace info about FRAME to stderr"""
    ## DEBUG:
    sys.stderr.write(f"{label}: {frame.f_code.co_name} {inspect.getfile(frame)}:{frame.f_lineno}\n")

class MezclaDebugger:
    """
    A class used to provide debugging functionality with formatted output.

    Methods:
    -------
    format(self, *args, arg_offset=0)
        Formats the given arguments and returns the formatted string

    Note:
    - See debug.trace_expr for various kwargs supported:
      delim, no_eol, max_len, etc.
    """

    _pairSeparator = "; "
    _lineWrapWidth = DEFAULT_LINE_WRAP_WIDTH
    _contextDelimiter = DEFAULT_CONTEXT_DELIMITER

    def __init__(
        self,
        prefix="",
        output_function=DEFAULT_OUTPUT_FUNCTION,
        arg_to_string_function=argument_to_string,
        include_context=False,
        context_abs_path=False,
        icecream_like=None,
    ):
        self.enabled = True
        self.prefix = prefix
        self.include_context = include_context
        self.output_function = output_function
        self.arg_to_string_function = arg_to_string_function
        self.context_abs_path = context_abs_path
        self.icecream_like = icecream_like

    def __call__(self, *args, arg_offset=0, indirect=False, **kwargs):
        """
        Formats the given arguments and prints the formatted string
        
        :raises ValueError: If the call frame cannot be accessed
        """
        # NOTE: this is not separated into a function
        # as to not generate another call frame
        call_frame = inspect.currentframe() 
        ## DEBUG: trace_frame(call_frame, "call_frame0")
        if call_frame is not None:
            # note: only go back one frame from this function
            go_back = indirect or (call_frame.f_code.co_name != "__call__")
            call_frame = call_frame.f_back
            ## DEBUG: trace_frame(call_frame, "call_frame1")
            if go_back and call_frame is not None:
                call_frame = call_frame.f_back
                ## DEBUG: trace_frame(call_frame, "call_frame2")
        if call_frame is None:
            raise ValueError("Cannot access the call frame")
        self.output_function(self._format(call_frame, arg_offset, *args, **kwargs))
        
        if not args:
            passthrough = None
        elif len(args) == 1:
            passthrough = args[0]
        else:
            passthrough = args
        return passthrough

    def format(self, *args, arg_offset=0, indirect=False, **kwargs):
        """
        Formats the given arguments and returns the formatted string,
        ignoring the first `arg_offset` arguments
        
        :raises ValueError: If the call frame cannot be accessed
        """
        ## TODO2: add levels_back param (e.g., 2 for gh.assertion)
        # NOTE: this is not separated into a function
        # as to not generate another call frame
        call_frame = inspect.currentframe()
        ## DEBUG: trace_frame(call_frame, "call_frame0")
        if call_frame is not None:
            # note: only go back one frame from this function
            go_back = indirect or (call_frame.f_code.co_name != "format")
            call_frame = call_frame.f_back
            ## DEBUG: trace_frame(call_frame, "call_frame1")
            if go_back and call_frame is not None:
                call_frame = call_frame.f_back
                ## DEBUG: trace_frame(call_frame, "call_frame2")
        if call_frame is None:
            raise ValueError("Cannot access the call frame")
        out = self._format(call_frame, arg_offset, *args, **kwargs)
        return out
    
    def get_context(self, call_frame: FrameType):
        """
        Returns the context of the given call frame
        """
        return self._format_context(call_frame)

    def _format(self, call_frame: FrameType, arg_offset: int, *args, **kwargs):
        """
        Formats the given arguments and returns the formatted string
        """
        
        ## OLD: prefix = pref if (pref := kwargs.get('prefix')) is not None else call_or_value(self.prefix)
        pref = kwargs.get('prefix') or kwargs.get('_prefix')
        prefix = pref if pref is not None else call_or_value(self.prefix)
        ## HACK: uses _prefix kwarg to avoid conflict with positional arg
        kwargs["_prefix"] = prefix

        context = self._format_context(call_frame)
        if self.icecream_like and not args:
            time = self._format_time()
            out = prefix + context + time
        else:
            if not self.include_context:
                context = ""
            out = self._format_args(call_frame, arg_offset, prefix, context, args, **kwargs)

        return out

    def _format_args(self, call_frame, arg_offset, prefix, context, args, **kwargs):
        """
        formats the arguments in `call_frame` with `prefix`,
        ignoring the first `arg_offset` arguments,
        optionally by adding `context` to the output
        """
        ## DEBUG: sys.stderr.write(f"in _format_args: {call_frame=}, {arg_offset=}, {prefix=}, {context=}, {args=}, {kwargs=}\n")
        call_node = Source.executing(call_frame).node
        ## DEBUG: sys.stderr.write(f"call_node: {ast.dump(call_node)}\n")
        if call_node is not None:
            source = Source.for_frame(call_frame)
            # Note: disables mypy error: "EnhancedAST" has no attribute "args" [attr-defined]
            ## DEBUG: sys.stderr.write(f"{call_node=} {call_node.args=} {source=}\n")
            sanitized_arg_strs = [
                source.get_text_with_indentation(arg)
                for arg in call_node.args[arg_offset:]      # type: ignore [attr-defined]
            ]
            ## DEBUG: sys.stderr.write(f"{sanitized_arg_strs=}\n")
        else:
            sys.stderr.write(f"Warning: unable to resolve call node: {args=}\n")
            sanitized_arg_strs = [_absent] * len(args)

        pairs = list(zip(sanitized_arg_strs, args))
        ## OLD: no_eol = kwargs.get('eol', False)
        ## DEBUG: sys.stderr.write(f"{pairs=}\n")

        ## OLD: out = self._construct_argument_output(prefix, context, pairs, no_eol=no_eol, **kwargs)
        out = self._construct_argument_output(prefix, context, pairs, **kwargs)
        return out

    def _construct_argument_output(self, prefix, context, pairs, **kwargs):
        ## OLD: def _construct_argument_output(self, prefix, context, pairs, no_eol=False, **kwargs):
        """
        Constructs the output string from the given pairs of arguments and values,
        """
        def arg_prefix(arg, delim='=') -> str:
            """Return ARG concatenated with DELIM"""
            return f"{arg}{delim}"
        def format_value(val, max_len):
            """Return up to MAX_LEN of VAL text, adding ... if truncated"""
            result = val
            if isinstance(max_len, int) and len(val) > max_len:
                result = val[:max_len + 1] + "..."
            return result

        pairs = [(arg, self.arg_to_string_function(val)) for arg, val in pairs]
        if "max_len" in kwargs:
            pairs = [(arg, format_value(val, kwargs["max_len"]))
                     for (arg, val) in pairs]
        pair_strs = [
            val if (is_literal(arg) or arg is _absent) else (arg_prefix(arg) + val)
            for arg, val in pairs
        ]
        suffix = kwargs.get('suffix', '')
        ## OLD: separator = sep if (sep := kwargs.get('sep')) else self._pairSeparator
        separator = sep if (sep := kwargs.get('delim')) else self._pairSeparator
        all_args_on_one_line = separator.join(pair_strs)
        multiline_args = len(all_args_on_one_line.splitlines()) > 1

        # note: context stuff is relic of icecream
        context_delimiter = self._contextDelimiter if context else ""
        all_pairs = prefix + context + context_delimiter + all_args_on_one_line
        ## BAD: first_line_too_long = len(all_pairs.splitlines()[0]) > self._lineWrapWidth
        first_line_too_long = (len(all_pairs.splitlines()[0]) > self._lineWrapWidth if all_pairs else False)

        if self.icecream_like and (multiline_args or first_line_too_long):
            if context:
                lines = [prefix + context] + [
                    format_pair(len(prefix) * " ", arg, value) for arg, value in pairs
                ]
            else:
                arg_lines = [format_pair("", arg, value) for arg, value in pairs]
                lines = prefix_first_line_indent_remaining(prefix, "\n".join(arg_lines))
        else:
            ## OLD: lines = [prefix + context + context_delimiter + all_args_on_one_line + suffix]
            ## OLD2: lines = "".join((v if v else "") for v in
            ##                       [prefix, context, context_delimiter, all_args_on_one_line, suffix])
            lines = [(v if v else "") for v in
                     [prefix, context, context_delimiter, all_args_on_one_line, suffix]]
        ## DEBUG: sys.stderr.write(f"{prefix=}\n{context=}\n{context_delimiter=}\n{all_args_on_one_line=}\n{suffix=}\n")

        no_eol = kwargs.get('no_eol', False)
        end = "" if no_eol else "\n"
        ## OLD: return end.join(lines)
        return "".join(lines) + end

    def _format_context(self, call_frame: FrameType) -> str:
        """
        Formats the context of the given call frame
        """
        filename, line_number, parent_function = self._get_context(call_frame)

        if parent_function != "<module>":
            parent_function = f"{parent_function}()"

        context = f"{filename}:{line_number} in {parent_function}"
        return context

    def _format_time(self) -> str:
        """
        Formats the current time
        """
        now = datetime.now()
        formatted = now.strftime("%H:%M:%S.%f")[:-3]
        return f" at {formatted}"

    def _get_context(self, call_frame):
        """
        Returns the context of the given call frame
        """
        frame_info = inspect.getframeinfo(call_frame)
        line_number = frame_info.lineno
        parent_function = frame_info.function

        filepath = (realpath if self.context_abs_path else basename)(
            frame_info.filename
        )
        return filepath, line_number, parent_function

#...............................................................................

## OLD:
## # def init():

intro = MezclaDebugger()

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    stderr_print(f"Warning: {basename(__file__)} not intended for direct invocation!")



#! /usr/bin/env python3
#
# Uses the Hugging Face API for machine translation (MT)
#
# Based on:
# - https://stackoverflow.com/questions/70043467/how-to-run-huggingface-helsinki-nlp-models
# - Hugging Face's NLP with Transformers text
#
# TODO3:
# - Address HuggingFace tip:
#   In order to maximize efficiency [on a GPU] please use a dataset. [instead of using pipelines sequentially]
#
# TODO2:
# - Decompose main using helper class(es): see ../template.py and ../simple_main_example.py.
#

"""Machine translation via Hugging Face

Examples:

echo "How now Bourne cow?" | FROM=en TO=es {script} -

USE_INTERFACE=1 {script} -

--model "BryanFalkowski/english-to-latin-v2" --from=en --to=la - <<<"The farmer is in the field"
"""
#
## TODO3: find a model supporting Latin to English
## {script} --model "BryanFalkowski/english-to-latin-v2" --from=la --to=en - <<<"Habemus Papam"
## NOTE: It should produce "We have a pope".

# Standard modules
## TODO: import json

# Intalled modules
# Note: done dynamically below

# Local modules
from mezcla import debug
from mezcla.main import Main, USE_PARAGRAPH_MODE
# TODO2: add new mezcla.hugging_face module for common stuff
import mezcla.examples.hugging_face_speechrec as hf_speechrec
from mezcla import misc_utils
from mezcla.my_regex import my_re
from mezcla import system
from mezcla import glue_helpers as gh

# Constants and Environment options
TL = debug.TL

FROM = system.getenv_text("FROM", "es")
TO = system.getenv_text("TO", "en")
SOURCE_LANG = system.getenv_text("SOURCE_LANG", FROM,
                                 "Source language")
TARGET_LANG = system.getenv_text("TARGET_LANG", TO,
                                 "Target language")
debug.assertion(SOURCE_LANG != TARGET_LANG)
SHOW_ELAPSED = system.getenv_bool("SHOW_ELAPSED", False,
                                  "Show elapsed time")
MAX_LENGTH = system.getenv_value(
    "MAX_LENGTH", None,
    description="Optional maximum length of tokens")

TEXT_ARG = "text"
FROM_ARG = "from"
TO_ARG = "to"
## TODO: ELAPSED_ARG = "elapsed-time"
TASK_ARG = "task"
MODEL_ARG = "model"
UI_ARG = "ui"
ROUND_ARG = "round"
TRANSLATION_TEXT = "translation_text"

USE_GPU = system.getenv_bool("USE_GPU", True, "Uses Torch on GPU if True")
MAX_LENGTH = system.getenv_int("MAX_LENGTH", 512, "Maximum Length of Tokens")
TEXT_FILE = system.getenv_text("TEXT_FILE", "-",
                               "Text file to translate")
USE_INTERFACE = system.getenv_bool("USE_INTERFACE", False,
                                   "Use web-based interface via gradio")

# Note: Round-trip translation: Translating text from one language to another and back to its original form
ROUND_TRIP = system.getenv_bool("ROUND_TRIP", False, 
                                "Perform round-trip translation")

#-------------------------------------------------------------------------------

def show_gpu_usage(trace_level=None):
    """Show nvidia usage if under GPU"""
    if trace_level is None:
        trace_level = 5
    if (hf_speechrec.TORCH_DEVICE == "GPU"):
        debug.code(trace_level, lambda: debug.trace(1, gh.run("nvidia-smi")))
    return

def translated_text(model_obj):
    """Get translation text for first entry in MODEL_OBJ"""
    return model_obj[0][TRANSLATION_TEXT] or ""

def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    
    # Show simple usage if --help given
    dummy_app = Main(description=__doc__.format(script=gh.basename(__file__)),
                     skip_input=False, manual_input=True,
                     boolean_options=[
                         ## TODO3: (ELAPSED_ARG, "Show elapsed time")],
                         (UI_ARG, "Invoke user interface"),
                         (ROUND_ARG, "Enable round-trip translation")
                         ],
                     text_options=[
                         (FROM_ARG, "Source language code"),
                         (TO_ARG, "Target language code"),
                         (TASK_ARG, "Translation task"),
                         (MODEL_ARG, "Model for translation"),
                         (TEXT_ARG, "Text to translate")])
    
    debug.trace_object(5, dummy_app)
    debug.assertion(dummy_app.parsed_args)
    text = dummy_app.get_parsed_option(TEXT_ARG)
    source_lang = dummy_app.get_parsed_option(FROM_ARG, SOURCE_LANG)
    target_lang = dummy_app.get_parsed_option(TO_ARG, TARGET_LANG)
    # Round-trip from argument
    round_trip = dummy_app.get_parsed_option(ROUND_ARG, ROUND_TRIP)
    use_interface = dummy_app.get_parsed_option(UI_ARG, USE_INTERFACE)
    
    MT_TASK = f"translation_{source_lang}_to_{target_lang}"                 # pylint: disable=invalid-name
    MT_MODEL = f"Helsinki-NLP/opus-mt-{source_lang}-{target_lang}"          # pylint: disable=invalid-name
    mt_task = dummy_app.get_parsed_option(TASK_ARG, MT_TASK)
    mt_model = dummy_app.get_parsed_option(MODEL_ARG, MT_MODEL)

    # Creating language models and tasks in reverse for round-trip translation
    if round_trip:
        MT_TASK_REVERSE = f"translation_{target_lang}_to_{source_lang}"                 # pylint: disable=invalid-name
        MT_MODEL_REVERSE = f"Helsinki-NLP/opus-mt-{target_lang}-{source_lang}"          # pylint: disable=invalid-name  
        mt_task_reverse = dummy_app.get_parsed_option(TASK_ARG, MT_TASK_REVERSE)
        mt_model_reverse = dummy_app.get_parsed_option(MODEL_ARG, MT_MODEL_REVERSE)

    # Get input file
    debug.trace_expr(5, dummy_app.input_stream, text, TEXT_FILE)
    text_file = TEXT_FILE
    if ((text is not None) or use_interface):
        pass
    elif text_file == "-":
        text_file = dummy_app.temp_file
        text = dummy_app.read_entire_input()
    else:
        text = system.read_file(text_file)

    # Load PyTorch and Hugging Face pipeline support
    # pylint: disable=import-outside-toplevel
    torch = hf_speechrec.init_torch_etc()
    debug.trace_expr(4, torch)
    if torch is None:
        import torch
        debug.trace_expr(4, torch)
    from transformers import pipeline   
    
    # Load Model
    device = torch.device(hf_speechrec.TORCH_DEVICE)
    debug.trace_expr(5, device)
    model = pipeline(task=mt_task, model=mt_model, device=device)

    # Create a model for reverse translation when ROUND_TRIP is true
    model_reverse = None
    if round_trip:
        model_reverse = pipeline(task=mt_task_reverse, model=mt_model_reverse)

    def translate(text, mt_model=None, reverse=False):
        """Translate TEXT from current source language to target
        Optionally translates in REVERSE or uses different MT_MODEL
        Note: works around issue with non-Helsinki-NLP models requiring source and target spec
        """
        ## TODO3: Convert into method (once main decomposed using class)
        if mt_model is None:
            mt_model = (model if not reverse else model_reverse)
        translation = [{TRANSLATION_TEXT: ""}]
        try:
            translation = mt_model(text, max_length=MAX_LENGTH)
        except:
            debug.trace_exception(6, "translate")
            debug.trace(5, "FYI: Retrying with explicit source and target")
            try:
                translation = mt_model(text, max_length=MAX_LENGTH,
                                       src_lang=source_lang, tgt_lang=target_lang)
            except:
                system.print_exception_info("translate")
        return translation

    # Pull up web interface if requested
    if use_interface:
        # TODO2: add language controls
        import gradio as gr             # pylint: disable=import-outside-toplevel
        pipeline_if = gr.Interface.from_pipeline(
            model,
            title="Machine translation (MT)",
            ## TODO2: subtitle=f"From: {FROM}; To: {TO}",
            description=f"From: {FROM}; To: {TO}",
            ## TODO3: examples=[...]); See example in hf_stable_diffusion.py.
            )
        pipeline_if.launch()

    # Otherwise, do translation and output
    else:
        split_regex = r"\n\s*\n" if USE_PARAGRAPH_MODE else "\n"
        # Note: Avoids "I'm sorry" bug when reading from stdin
        segments = my_re.split(split_regex, text)
        debug.trace_expr(6, segments)
        if segments[-1] == "":
            segments = segments[:-1]
        for s, segment in enumerate(segments):
            debug.trace_expr(3, s, segment)
            if not segment.strip():
                continue
            try:
                translation_reverse_text = translation_round_text = ""

                # Translation Level I (FROM -> TO)
                ## OLD: translation = model(segment, max_length=MAX_LENGTH)
                translation = translate(segment)
                translation_text = translated_text(translation)
                debug.trace_expr(3, translation_text)
                
                # Round-Trip translation uses the reverse model to re-translate back to original form
                ## TODO2: drop intermediate translation in round-trip support
                ## NOTE: Only needs to go from source to target and then target back to source:
                ##    src-original -> tgt-translated -> src-round-trip
                if round_trip:
                    # Translation Level II (TO -> FROM_AUX)
                    ## OLD: translation_reverse = model_reverse(translation_text, max_length=MAX_LENGTH)
                    translation_reverse = translate(segment, reverse=True)
                    translation_reverse_text = translated_text(translation_reverse)

                    # Translation Level III (FROM_AUX -> TO)
                    ## OLD: translation_round = model(translation_reverse_text, max_length=MAX_LENGTH)
                    translation_round = translate(translation_reverse_text)
                    translation_round_text = translated_text(translation_round)

                debug.assertion(isinstance(translation, list)
                                and (TRANSLATION_TEXT in translation[0]))
                ## OLD:
                ## translation_reverse_text = translation_reverse_text if round_trip else ''
                ## translation_round_text = translation_round_text if round_trip else ''
                
                # For round trip translation, print all possible translations along with their language code
                if round_trip:
                    print(f"\nOriginal      ({FROM}):\n{segment}")
                    print(f"\nTranslate     ({TO}):\n{translation_text}")
                    print(f"\nOriginal  [R]   ({FROM}):\n{translation_reverse_text}")
                    print(f"\nTranslate [R]  ({TO}):\n{translation_round_text}")
                    round_trip_diff_original =  (segment != translation_reverse_text)
                    round_trip_diff_translate = (translation_round_text != translation_text)
                    round_trip_difference = round_trip_diff_original or round_trip_diff_translate
                    print(f"\nDifference in Translation: {round_trip_difference}\n")
                    
                    # If there is any difference during round-trip translation, print the difference
                    if round_trip_difference:
                        print("="*40)
                        print(f"\nDifferences in Original ({FROM}):")
                        print(f"{misc_utils.string_diff(segment, translation_reverse_text) if round_trip_diff_original else None}")
                        print(f"Differences in Translated ({TO}):")
                        print(f"{misc_utils.string_diff(translation_text, translation_round_text) if round_trip_diff_translate else None}")
                else:
                    print(translation_text)
            except:
                system.print_exception_info("translation")
            
            show_gpu_usage()

    # Wrap up
    show_gpu_usage()
    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    elapsed = misc_utils.time_function(main)
    if SHOW_ELAPSED:
        print(f"Elapsed time: {system.round_as_str(elapsed)}ms")



#! /usr/bin/env python3
#
# Dummy module for test_mezcla_to_standard.py
#

"""Dummy module for testing purposes"""

def func(*_args):
    """Dummy function with variable number of _ARGS"""
    return
func2 = func3 = func4 = func1 = func



#! /usr/bin/env python3
# TODO: # -*- coding: utf-8 -*-
## TODO: handle case when env installed elsewhere (e.g., maldito mac)
## #! env python
# 
# TODO what the script does (detailed)
#
## TODO: see example/template.py for simpler version suitable for cut-n-paste from online examples
#

"""
TODO: what module does (brief)

Sample usage:
   echo $'TODO:task1\\nDONE:task2' | {script} --TODO-arg --
"""

# Standard modules
from typing import Optional

# Installed modules
## TODO: import numpy

# Local modules
# TODO: def mezcla_import(name): ... components = eval(name).split(); ... import nameN-1.nameN as nameN
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla.my_regex import my_re
from mezcla import system
## TODO:
## from mezcla import data_utils as du
##
## Optional:
## # Increase trace level for regex searching, etc. (e.g., from 6 to 7)
## my_re.TRACE_LEVEL = debug.QUITE_VERBOSE
debug.trace(5, f"global __doc__: {__doc__}")
debug.assertion(__doc__)

## TODO: Constants for switches omitting leading dashes (e.g., DEBUG_MODE = "debug-mode")
## Note: Run following in Emacs to interactively replace TODO_ARG with option label
##    M-: (query-replace-regexp "todo\\([-_]\\)arg" "arg\\1name")
## where M-: is the emacs keystroke short-cut for eval-expression.
TODO_ARG = "TODO-arg"
## TEXT_ARG = "text-arg"
## ALT_FILENAME = "alt_filename"

# Constants
TL = debug.TL

# Environment options
# Note: These are just intended for internal options, not for end users.
# It also allows for enabling options in one place rather than four
# (e.g., [Main member] initialization, run-time value, and argument spec., along
# with string constant definition).
# WARNING: To minimize environment comflicts with other programs make the names
# longer such as two or more tokens (e.g., "FUBAR" => "FUBAR_LEVEL").
#
TODO_FUBAR = system.getenv_bool(
    "TODO_FUBAR", False,
    description="TODO:Fouled Up Beyond All Recognition processing")
## TODO
## X = system.getenv_int(
##     "X", X,
##     description="X")


## TODO: Use helper class for main logic
## class Helper:
##     """TODO: class for doing ..."""
## 
##     def __init__(self, ...) -> None:
##         """Initializer: ..."""
##         debug.trace(TL.VERBOSE, f"Helper.__init__(): self={self}")
##         self.TODO = None
##         debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")
## 
##     def process(self, ...) -> None:
##         """TODO: ..."""
##         # TODO: flesh out
##

class Script(Main):
    """Input processing class"""
    # TODO: -or-: """Adhoc script class (e.g., no I/O loop, just run calls)"""
    ## TODO: class-level member variables for arguments (avoids need for class constructor)
    todo_arg: Optional[bool] = False
    ## text_arg = ""

    # TODO: add class constructor if needed for non-standard initialization
    ## WARNING: For Script classes involving complex logic, it is best to use helper classes,
    ## as done in show_bert_representation.py.
    ## NOTE: Such class decomposition is also beneficial for unit tests.
    #
    ## def __init__(self, *args, **kwargs):
    ##     debug.trace_expr(TL.VERBOSE, self, args, kwargs, delim="\n\t", prefix="in {self.__class__.__name__}.__init__({args}, {kwargs})")
    ##     super().__init__(*args, **kwargs)

    def setup(self) -> None:
        """Check results of command line processing"""
        debug.trace(TL.VERBOSE, f"Script.setup(): self={self}")
        ## TODO: extract argument values
        self.todo_arg = self.get_parsed_option(TODO_ARG, self.todo_arg)
        ## TODO:
        ## self.text_arg = self.get_parsed_option(TEXT_ARG, self.text_arg)
        ## self.alt_filename = self.get_parsed_argument(ALT_FILENAME)
        ## OLD:
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")
        ## TEST: debug.trace_object(5, self, label=f"{self.__init__.__qualname__.split('.')[0]} instance")

    def process_line(self, line) -> None:
        """Processes current line from input"""
        debug.trace_fmtd(TL.QUITE_DETAILED, "Script.process_line({l})", l=line)
        # TODO: flesh out
        if self.todo_arg and "TODO" in line:
            print(f"arg1 line ({self.line_num}): {line}")
        else:
            debug.trace(3, f"Ignoring line ({self.line_num}): {line}")
        ## TODO: regex pattern matching
        ## elif my_re.search(self.text_arg, line):
        ##     print("arg2 line: %s" % line)

    ## TODO: if no input prococessed, customize run_main_step instead
    ## and specify skip_input below. (Use skip_input=False for filename support.)
    ##
    ## def run_main_step(self) -> None:
    ##     """Main processing step (n.b., assumes self.manual_input)"""
    ##     debug.trace(5, f"Script.run_main_step(): self={self}")
    ##     ...
    ##     ## TODO: data = self.read_entire_input()
    #@     ...
    ##

    ## TODO:
    ## def wrap_up(self) -> None:
    ##     """Do final processing"""
    ##     debug.trace(6, f"Script.wrap_up(); self={self}")
    ##     # ...

def main() -> None:
    """Entry point"""
    app = Script(
        description=__doc__.format(script=gh.basename(__file__)),
        # Note: if not manual_input, line-by-line processing is done via process_line;
        # otherwise, run_main_step is used. Use skip_input to disable filename argument.
        skip_input=False,
        manual_input=False,
        ## TODO (specify auto_help such as when manual_input set):
        ## # Note: shows brief usage if no arguments given
        ## auto_help=True,
        ## -or-: # Disable inference of --help argument
        ## auto_help=False,
        ## TODO: specify options and (required) arguments
        boolean_options=[(TODO_ARG, "TODO-desc--currently greps for TODO")],
        ## TODO
        ## Note: FILENAME is default argument unless skip_input
        ## positional_arguments=[ALT_FILENAME], 
        ## text_options=[(TEXT_ARG, "TODO-desc")],
        ## Note: Following added for indentation float options not common (TODO: remove?)
        float_options=None)
    app.run()
    # Make sure no TODO_vars above (i.e., in namespace)
    debug.assertion(not any(my_re.search(r"^TODO_", m, my_re.IGNORECASE)
                            for m in dir(app)))

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    debug.trace(5, f"module __doc__: {__doc__}")
    debug.assertion("TODO:" not in __doc__)
    main()



#! /usr/bin/env python3
#
# compute_tfidf.py: compute Term Frequency Inverse Document Frequency (TF-IDF)
# for a collection of documents. See https://en.wikipedia.org/wiki/Tf-idf.
#
# TODO:
# - Show examples.
# - Have option for producing document/term matrix.
# - Add option to omit TF, IDF and IDF fields (e.g., in case just interested in frequency counts).
# - Add option to put term at end, so numeric fields are aligned.
# - -and/or- Have ** max-output-term-length option and right pad w/ spaces.
# - See if IDF calculation should for 0 if just one document occurrence.
# - Reconcile with ngram_tfidf.py (e.g., overlap here with subsumption there).
# - Add simple example(s) to help.
#
# Note:
# - This script is just for running tfidf over text files.
# - See ngram_tfdif.py for a wrapper class around tfidf for use in applications
#   like Visual Diff Search that generate the text dynamically.
#


"""Compute Term Frequency/Inverse Document Frequency (TF-IDF) for a set of documents"""

# Standard packages
import csv
import os
import re
import sys

# Installed packages
# TODO: require version 1.1 with TPO hacks
from mezcla import tfidf
from mezcla.tfidf import MIN_NGRAM_SIZE, MAX_NGRAM_SIZE
from mezcla.tfidf.corpus import Corpus as tfidf_corpus
from mezcla.tfidf.preprocess import Preprocessor as tfidf_preprocessor

# Local packages
from mezcla import debug
from mezcla import system
from mezcla.system import PRECISION
from mezcla.text_utils import make_fixed_length

# Determine environment-based options
DEFAULT_NUM_TOP_TERMS = system.getenv_int("NUM_TOP_TERMS", 10)
## OLD:
## MAX_NGRAM_SIZE = system.getenv_int("MAX_NGRAM_SIZE", 1)
## MIN_NGRAM_SIZE = system.getenv_int("MIN_NGRAM_SIZE", MAX_NGRAM_SIZE)
USE_NGRAM_SMOOTHING = system.getenv_boolean("USE_NGRAM_SMOOTHING", False)

## OLD: IDF_WEIGHTING = system.getenv_text("IDF_WEIGHTING", "basic")
DEFAULT_IDF_WEIGHTING = 'smooth' if USE_NGRAM_SMOOTHING else 'basic'
IDF_WEIGHTING = system.getenv_text("IDF_WEIGHTING", DEFAULT_IDF_WEIGHTING)
## OLD: TF_WEIGHTING = system.getenv_text("TF_WEIGHTING", "basic")
## TODO3: DEFAULT_IF_WEIGHTING = 'smooth' if USE_NGRAM_SMOOTHING else 'log'
DEFAULT_TF_WEIGHTING = 'basic'
TF_WEIGHTING = system.getenv_text("TF_WEIGHTING", DEFAULT_TF_WEIGHTING)
DELIMITER = system.getenv_text("DELIMITER", ",")
CORPUS_DUMP = system.getenv_value("CORPUS_DUMP", None,
                                  "Filename for corpus dump")
PRUNE_SUBSUMED_TERMS = system.getenv_bool("PRUNE_SUBSUMED_TERMS", False)
PRUNE_OVERLAPPING_TERMS = system.getenv_bool("PRUNE_OVERLAPPING_TERMS", False)
SKIP_STEMMING = system.getenv_bool("SKIP_STEMMING", False,
                                   "Skip word stemming (via Snowball)")
INCLUDE_STEMMING = not SKIP_STEMMING
STEMMER_LANGUAGE = system.getenv_value("STEMMER_LANGUAGE", None,
                                       "Language for stemming and stop words--not recommended")
LANGUAGE = (STEMMER_LANGUAGE or "")
TAB_FORMAT = system.getenv_bool("TAB_FORMAT", False,
                                 "Use tab-delimited format--facilitates spreadsheet import")
TERM_WIDTH = system.getenv_int("TERM_WIDTH", 32,
                               "Width of term column in output")
SCORE_WIDTH = system.getenv_int("SCORE_WIDTH", PRECISION + 6,
                                "Width of each score column in output (e.g., up to 12 for default precision of 6 as in 1.855712e-03)")
MAX_FIELD_SIZE = system.getenv_int(
    "MAX_FIELD_SIZE", -1,
    desc="Overide for default max field size (128k)")

# Option names and defaults
NGRAM_SIZE_OPT = "--ngram-size"
NUM_TOP_TERMS_OPT = "--num-top-terms"
SHOW_SUBSCORES = "--show-subscores"
SHOW_FREQUENCY = "--show-frequency"
CSV = "--csv"
TSV = "--tsv"
TEXT = "--text"

#...............................................................................

def show_usage_and_quit():
    """Show command-line usage for script and then exit"""
    # TODO: make 
    usage = """
Usage: {prog} [options] file1 [... fileN]

Options: [--help] [{ngram_size_opt}=N] [{top_terms_opt}=N] [{subscores}] [{frequencies}] [{csv} | {tsv} | {text}]

Notes:
- Derives TF-IDF for set of documents, using single word tokens (unigrams), by default. 
- By default, the document ID is the position of the file on the command line (e.g., N for fileN above). The document text is the entire file.
- However, with {csv}, the document ID is taken from the first column, and the document text from the second columns (i.e., each row is a distinct document).
- With {text}, the document ID is taken from the line number.
- Use following environment options:
      DEFAULT_NUM_TOP_TERMS ({default_topn})
      MIN_NGRAM_SIZE ({min_ngram_size})
      MAX_NGRAM_SIZE ({max_ngram_size})
      TF_WEIGHTING ({tf_weighting}): {{log, norm_50, binary, basic, freq}}
      IDF_WEIGHTING ({idf_weighting}): {{smooth, max, prob, basic, freq}}
""".format(prog=sys.argv[0], ngram_size_opt=NGRAM_SIZE_OPT, top_terms_opt=NUM_TOP_TERMS_OPT, subscores=SHOW_SUBSCORES, frequencies=SHOW_FREQUENCY, default_topn=DEFAULT_NUM_TOP_TERMS, min_ngram_size=MIN_NGRAM_SIZE, max_ngram_size=MAX_NGRAM_SIZE, tf_weighting=TF_WEIGHTING, idf_weighting=IDF_WEIGHTING, csv=CSV, tsv=TSV, text=TEXT)
    print(usage)
    sys.exit()


def get_suffix1_prefix2(subterms1, subterms2):
    """returns any suffix of SUBTERMS1 list that is a prefix of SUBTERMS2 list"""
    # EX: get_suffix1_prefix2(["my", "dog"], ["dog", "has"]) => ["dog"]
    # EX: get_suffix1_prefix2(["a", "b", "c"], ["b", "d"]) => []
    # TODO: support string arguments (e.g., by splitting on whitespace)
    prefix_len = 0
    for subterm1 in subterms1:
        if ((subterm1 in subterms2) and (subterms2.index(subterm1) == prefix_len)):
            prefix_len += 1
        else:
            prefix_len = 0
    prefix = (subterms1[-prefix_len:] if prefix_len else[])
    debug.trace_fmt(7, "get_suffix1_prefix2({st1}, {st2}) => {p}", st1=subterms1, st2=subterms2, p=prefix)
    return prefix
    

def terms_overlap(term1, term2):
    """Whether TERM1 and TERM1 overlap (and the overlapping text if so)
    Note: The overlap must occur at word boundaries
    """
    # EX: terms_overlap("ACME Rocket Research", "Rocket Research Labs") => "Rocket Research"
    # EX: not terms_overlap("Rocket Res", "Rocket Research")
    # TODO: put in text_utils
    subterms1 = term1.strip().split()
    subterms2 = term2.strip().split()
    overlap = ""
    if system.intersection(subterms1, subterms2):
        # pylint: disable=arguments-out-of-order
        overlap = " ".join((get_suffix1_prefix2(subterms1, subterms2) or get_suffix1_prefix2(subterms2, subterms1)))
    debug.trace_fmt(6, "terms_overlap({t1}, {t2}) => {o}", t1=term1.strip(), t2=term2.strip(), o=overlap)
    return overlap


def is_subsumed(term, terms, include_overlap=PRUNE_OVERLAPPING_TERMS):
    """Whether TERM is subsumed by another term in TERMS, accounting for overlapping terms if INCLUDE_OVERLAP
    Note: subsumption is based on string matching not token matching unlike the overlap check.
    """
    # EX: is_subsumed("White House", ["The White House", "Congress", "Supreme Court"])
    # EX: is_subsumed("White House", ["White Houses"])
    # TODO: Enforce word boundaries as with terms_overlap
    subsumed_by = [subsuming_term for subsuming_term in terms
                   if (((term in subsuming_term) 
                        or (include_overlap and terms_overlap(term, subsuming_term)))
                       and (term != subsuming_term))]
    subsumed = any(subsumed_by)
    debug.trace_fmt(6, "is_subsumed({t}) => {r}; subsumed by: {sb}",
                    t=term, r=subsumed, sb=subsumed_by)
    return subsumed

#...............................................................................

def main():
    """Entry point for script"""
    args = sys.argv[1:]
    ## NOTE: debug now shows args
    debug.trace_fmtd(4, "main()")
    debug.trace_fmtd(4, "os.environ={env}", env=os.environ)

    # Parse command-line arguments
    # TODO2: rework via Main (see examples/template.py)
    i = 0
    max_ngram_size = MAX_NGRAM_SIZE
    num_top_terms = DEFAULT_NUM_TOP_TERMS
    show_subscores = False
    show_frequency = False
    csv_file = False
    is_text = False
    global DELIMITER
    while ((i < len(args)) and args[i].startswith("-")):
        option = args[i]
        debug.trace_fmtd(5, "arg[{i}]: {opt}", i=i, opt=option)
        if (option == "--help"):
            show_usage_and_quit()
        elif (option == NGRAM_SIZE_OPT):
            i += 1
            max_ngram_size = int(args[i])
        elif (option == NUM_TOP_TERMS_OPT):
            i += 1
            num_top_terms = int(args[i])
        elif (option == SHOW_SUBSCORES):
            show_subscores = True
        elif (option == SHOW_FREQUENCY):
            show_frequency = True
            # Make sure TF-IDF package supports occurrence counts for TF
            tfidf_version = 1.0
            try:
                # Note major and minor revision values assumed to be integral
                major_minor = re.sub(r"^(\d+\.\d+).*", r"\1", tfidf.__version__)
                tfidf_version = float(major_minor)
            except:
                system.print_stderr("Exception in main: " + str(sys.exc_info()))
            assert(tfidf_version >= 1.2)
        elif (option == CSV):
            csv_file = True 
        elif (option == TSV):
            csv_file = True
            DELIMITER = "\t"
        elif (option == TEXT):
            csv_file = True
            is_text = True
            DELIMITER = "\xFF"
        else:
            sys.stderr.write("Error: unknown option '{o}'\n".format(o=option))
            show_usage_and_quit()
        i += 1
    args = args[i:]
    if (len(args) < 1):
        system.print_stderr("Error: missing filename(s)\n")
        show_usage_and_quit()
    if ((len(args) < 2) and (not csv_file) and (not show_frequency)):
        ## TODO: only issue warning if include-frequencies not specified
        system.print_stderr("Warning: TF-IDF not relevant with only one document")

    # Initialize Tf-IDF module
    debug.assertion(not re.search(r"^en(_\w+)?$", LANGUAGE, re.IGNORECASE))
    # Note: disables stemming via no-op lambda by default
    stemmer_fn = None if INCLUDE_STEMMING else (lambda x: x)
    my_pp = tfidf_preprocessor(language=LANGUAGE, gramsize=max_ngram_size, min_ngram_size=MIN_NGRAM_SIZE, all_ngrams=False, stemmer=stemmer_fn)
    corpus = tfidf_corpus(gramsize=max_ngram_size, min_ngram_size=MIN_NGRAM_SIZE, all_ngrams=False, preprocessor=my_pp)

    # Overide the maxium field size if specified
    if MAX_FIELD_SIZE > -1:
        old_limit = csv.field_size_limit()
        debug.assertion(MAX_FIELD_SIZE > old_limit)
        csv.field_size_limit(MAX_FIELD_SIZE)
        debug.trace(4, f"Set max field size to {MAX_FIELD_SIZE}; was {old_limit}")

    # Process each of the arguments
    doc_filenames = {}
    for i, filename in enumerate(args):
        # If CSS file, treat each row as separate document, using ID from first column and data from second
        if csv_file:
            text_col = 0 if is_text else 1
            with system.open_file(filename) as fh:
                csv_reader = csv.reader(iter(fh.readlines()), delimiter=DELIMITER, quotechar='"')
                # TODO: skip over the header line
                line = 0
                for r, row in enumerate(csv_reader):
                    debug.trace_fmt(6, "{l}: {r}", l=line, r=row)
                    doc_id = str(r + 1) if is_text else row[0]
                    try:
                        doc_text = system.from_utf8(row[text_col])
                    except:
                        debug.trace_fmt(5, "Exception processing line {l}", l=line)
                        doc_text = ""
                    ## TODO: use defaultdict-type hash
                    if doc_id not in corpus:
                        corpus[doc_id] = ""
                    else:
                        ## TODO: corpus[doc_id] += " "
                        corpus[doc_id] = (corpus[doc_id].text + " ")
                    # Appends text to corpus document
                    ## TODO: corpus[doc_id] += doc_text
                    corpus[doc_id] = (corpus[doc_id].text + doc_text)
                    ## OLD: doc_filenames[doc_id] = filename + ":" + str(i + 1)
                    doc_filenames[doc_id] = f"{filename}:{r + 1}"
                    line += 1
        # Otherwise, treat entire file as document and use command-line position as the document ID
        else:
            doc_id = str(i + 1)
            doc_text = system.read_entire_file(filename)
            corpus[doc_id] = doc_text
            doc_filenames[doc_id] = filename
    debug.trace_object(7, corpus, "corpus")
    if CORPUS_DUMP:
        system.save_object(CORPUS_DUMP, corpus)

    # Derive headers
    headers = ["term"]
    if show_frequency:
        headers += ["TFreq", "DFreq"]
    if show_subscores:
        headers += ["TF", "IDF"]
    headers += ["TF-IDF"]
    debug.assertion(headers[0] == "term")

    # Output the top terms per document with scores
    # TODO: change the IDF weighting
    for doc_id in corpus.keys():        # pylint: disable=consider-using-dict-items
        print("{id} [{filename}]".format(id=doc_id, filename=doc_filenames[doc_id]))
        if TAB_FORMAT:
            print("\t".join(headers))
        else:
            term_header_spec = make_fixed_length(headers[0], TERM_WIDTH)
            other_header_spec = " ".join([make_fixed_length(h, SCORE_WIDTH) for h in headers[1:]])
            print(term_header_spec + " " + other_header_spec)

        # Get ngrams for document and calculate overall score (TF-IDF).
        # Then print each in tabular format (e.g., "et al   0.000249")
        top_term_info = corpus.get_keywords(document_id=doc_id,
                                            idf_weight=IDF_WEIGHTING,
                                            limit=num_top_terms)
        # Optionally limit the result to terms that don't overlap with higher weighted ones.
        # TODO: Allow overlap if the terms occur in different parts of the document.
        if PRUNE_SUBSUMED_TERMS:
            top_terms = [term_info.ngram.strip() for term_info in top_term_info]
            top_term_info = [ti for (i, ti) in enumerate(top_term_info) 
                             if (not is_subsumed(ti.ngram, top_terms[0: i]))]
        for (term, score) in [(ti.ngram, ti.score)
                              for ti in top_term_info if ti.ngram.strip()]:
            # Get scores including component values (e.g., IDF)
            # TODO: don't round the frequency counts (e.g., 10.000 => 10)
            scores = []
            if show_frequency:
                scores.append(corpus[doc_id].tf_freq(term))
                scores.append(corpus.df_freq(term))
            if show_subscores:
                scores.append(corpus[doc_id].tf(term, tf_weight=TF_WEIGHTING))
                # TODO; idf_weight=TDF_WEIGHTING
                scores.append(corpus.idf(term))
            scores.append(score)

            # Print term and rounded scores
            if TAB_FORMAT:
                print(term + "\t" + "\t".join(map(system.round_as_str, scores)))
            else:
                rounded_scores = [make_fixed_length(system.round_as_str(s), SCORE_WIDTH) for s in scores]
                term_spec = make_fixed_length(system.to_utf8(term), TERM_WIDTH)
                score_spec = " ".join(rounded_scores)
                print(term_spec + " " + score_spec)
                ## TODO: debug.assertion((len(rounded_scores) * PRECISION) < len(score_spec) < (len(rounded_scores) * (1 + SCORE_WIDTH)))
        print("")

    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#! /usr/bin/env python3
#
# Test(s) for ../my_regex.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_my_regex.py
#

"""Tests for my_regex module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import debug

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.my_regex as THE_MODULE

# Constants
MEZCLA_REGEX = "M[e]zcl[a]"


class TestMyRegex(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    my_re = THE_MODULE.my_re            # TODO3: make global to cut down self usages

    def setUp(self):
        """Performs test setup with capsys disabled
           Note: This works around quirk with pytest stderr capturing"""
        try:
            with self.capsys.disabled():
                debug.trace(6, f"TestIt.setUp(); self={self}")
                super().setUp()
                debug.trace_current_context(level=debug.QUITE_DETAILED)
        except:
            # note: trace level high so as not to affect normal testing
            debug.trace_exception(7, "TestMyRegex.setUp")
            super().setUp()
   
    def helper_my_regex(self, regex, text, is_match=0):
        """Helper functions for my_regex"""
        ## TODO2: rename as get_regex_search_result
        if is_match:
            self.my_re.match(regex, text, 0)
        else:
            self.my_re.search(regex, text)
        output = self.my_re.get_match()
        return output
    
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_search(self):
        """Ensure search() works as expected"""
        debug.trace(4, f"test_search(); self={self}")
        text = "The quick brown fox jumps over the lazy dog."
        regex = r"\w{5}"
        output = self.helper_my_regex(regex, text)
        assert (output.group() == "quick" and output.span() == (4, 9))
    
    def test_search_alt(self):
        """Test search()"""
        debug.trace(4, f"test_search(); self={self}")
        self.do_assert(not self.my_re.search(MEZCLA_REGEX, "EZC"))
        debug.assertion("E" not in MEZCLA_REGEX)
        debug.assertion("e" in MEZCLA_REGEX)
        self.my_re.search(MEZCLA_REGEX, "EZC", flags=re.IGNORECASE)
        self.my_re.search(MEZCLA_REGEX, "ezc")
        self.do_assert(self.my_re.search_text == "ezc")

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_match(self):
        """Ensure match() works as expected"""
        debug.trace(4, f"test_match(); self={self}")
        text = "1 kiss is all takes."
        regex = r"\d+"
        output = self.helper_my_regex(regex, text, is_match=1)
        assert(output.group() == "1" and output.span() == (0, 1))
        ## TODO: return the matched value (solved: use group() after my_re.get_match())
        debug.assertion("M" in MEZCLA_REGEX)
        debug.assertion("m" not in MEZCLA_REGEX)
        self.do_assert(not self.my_re.match(MEZCLA_REGEX, "MEZC"))
        self.my_re.match(MEZCLA_REGEX, "MEZC", flags=re.IGNORECASE)
        self.my_re.match(MEZCLA_REGEX, "Mezc")
        self.do_assert(self.my_re.search_text == "Mezc")

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_get_match(self):
        """Ensure get_match() works as expected"""
        debug.trace(4, f"test_get_match(); self={self}")
        # get_match() returns the last the result of match
        text = "333 little birds"
        regex = r"\d+"
        output = self.helper_my_regex(regex, text, is_match=1)
        assert isinstance(output, re.Match)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_group(self):
        """Ensure group() works as expected"""
        debug.trace(4, f"test_group(); self={self}")
        text = "three, 7, eight"
        regex = r"\w{5},"
        self.my_re.search(regex, text, 0)
        output = self.my_re.group(0)
        assert (output == 'three,')

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_groups(self):
        """Ensure groups() works as expected"""
        debug.trace(4, f"test_groups(); self={self}")
        text = "John Doe: 30 years old, Jane Smith: 25 years old"
        regex = r"(\w+\s\w+): (\d+) years"
        self.my_re.search(regex, text)
        output = self.my_re.groups()
        assert(output == ('John Doe', '30'))

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_grouping(self):
        """Ensure grouping() works as expected"""
        debug.trace(4, f"test_grouping(); self={self}")
        text = "John Doe: 30 years old, Jane Smith: 25 years old"
        regex = r"(\w+\s\w+): (\d+) years"
        self.my_re.search(regex, text)
        output = self.my_re.grouping()
        assert(output == ('John Doe', '30'))

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_start(self):
        """Ensure start() works as expected"""
        debug.trace(4, f"test_start(); self={self}")
        text = "three little birds"
        regex = r"\w{5}"
        self.my_re.search(regex, text)
        output = self.my_re.start()
        # start() returns starting index
        assert(output == 0)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_end(self):
        """Ensure end() works as expected"""
        debug.trace(4, f"test_end(); self={self}")
        text = "three big birds"
        regex = r"\w{5}"
        self.my_re.search(regex, text)
        output = self.my_re.end()
        # start() returns ending index
        assert(output == 5)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_sub(self):
        """Ensure sub() works as expected"""
        debug.trace(4, f"test_sub(); self={self}")
        text = "The quick brown fox jumps over the lazy dog."
        regex = r"\w{4}"
        replacement = "****"
        output_sample = "The ****k ****n fox ****s **** the **** dog."
        output = self.my_re.sub(pattern=regex, string=text, replacement=replacement)
        assert(output == output_sample)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_span(self):
        """Ensure span() works as expected"""
        debug.trace(4, f"test_span(); self={self}")
        text = "The quick brown fox jumps over the lazy dog."
        regex = r"\w{4}"
        self.my_re.search(regex, text)
        output = self.my_re.span()
        assert(output == (4, 8))
    
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_split(self):
        """Ensure split() works as expected"""
        debug.trace(4, f"test_split(); self={self}")
        text = "three,little,birds"
        regex = r","
        output = self.my_re.split(pattern=regex, string=text)
        assert(output == text.split(","))

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_findall(self):
        """Ensure findall() works as expected"""
        debug.trace(4, f"test_findall(); self={self}")
        text = "There are 32768 possible combinations, with 256 other combinations and 0 impossible combinations."
        regex = r"\d+"
        output = self.my_re.findall(pattern=regex, string=text)
        assert(output == ['32768', '256', '0'])

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_escape(self):
        """Ensure escape() works as expected"""
        debug.trace(4, f"test_escape(); self={self}")
        text = "foo*bar"
        output = self.my_re.escape(text)
        assert(r"\*" in output)

    def check_f_string(self, expected=True):
        """Check that f-string warning is ISSUED"""
        debug.trace(4, f"check_f_string({expected})")
        self.my_re.search("{fubar}", "foobar")
        captured_stderr = self.get_stderr()
        debug.trace_expr(4, captured_stderr, max_len=4096)
        has_warning = bool(self.my_re.search("Warning:.*f-string", captured_stderr))
        self.do_assert(has_warning == expected)
        
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_f_string_warning(self):
        """Ensure warning given about f-string like regex"""
        # Make sure shown by default
        debug.trace(4, f"in test_f_string(); self={self}")
        self.monkeypatch.setattr(THE_MODULE, 'REGEX_WARNINGS', True)
        self.check_f_string(expected=True)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_no_f_string_warning(self):
        """Make sure f-string warning can be disabled"""
        debug.trace(4, f"in test_no_f_string_warning(); self={self}")
        self.monkeypatch.setattr(THE_MODULE, 'REGEX_WARNINGS', False)
        self.check_f_string(expected=False)

    def test_simple_regex(self):
        """"Test regex search with capturing"""
        debug.trace(4, f"test_simple_regex(); self={self}")
        self.do_assert(self.my_re.search(r"(\w+)\W+(\w+)", ">scrap ~!@\n#$ yard<",
                                         re.MULTILINE))
        self.do_assert(self.my_re.group(1) == "scrap")
        self.do_assert(self.my_re.group(2) == "yard")
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_pre_and_post_match(self):
        """Test pre/post_match() functions"""
        debug.trace(4, f"test_pre_and_post_match(); self={self}")
        self.my_re.search(r"[dD]ef", "abc_def_ghi")
        self.do_assert(self.my_re.pre_match() == "abc_")
        self.do_assert(self.my_re.post_match() == "_ghi")

    def check_pattern_helper(self, regex, expect_has_warning):
        """Make sure check_pattern warning for REGEX matches EXPECTED_HAS_WARNING"""
        # Note: Used in global test below, using separate tests due to capsys quirks.
        ## DEBUG: debug.trace(4, f"check_pattern_helper{(self, regex, expect_has_warning)}")
        ## TODO3: self.capsys.disabled() ...
        self.my_re.check_pattern(regex)
        captured_stderr = self.get_stderr()
        actual_has_warning = bool(self.my_re.search("Warning", captured_stderr))
        assert actual_has_warning == expect_has_warning


@pytest.mark.xfail                   # TODO: remove xfail
@pytest.mark.parametrize(
    ## TODO3: use unittest_parametrize (see test_mezcla_to_standard.py and https://pypi.org/project/unittest-parametrize)
    "regex, expect_warning",
    [
        ("{regex_var}", True),
        (b"{binary_regex_var}", True),
        ("regex_text", False),
        (b"binary_regex_text", False),
    ])
def test_check_pattern(regex, expect_warning, capsys):
    """Ensure check_pattern issues warning for REGEX if EXPECT_WARNING"""
    try:
        with capsys.disabled():
            debug.trace(4, f"test_check_pattern{(regex, expect_warning)}")
            debug.trace_expr(5, capsys)
            test_inst = TestMyRegex()
        test_inst.capsys = capsys
        test_inst.check_pattern_helper(regex, expect_warning)
    except AssertionError:
        debug.trace_exception(7, "test_check_pattern [assertion]")
        raise
    except:
        debug.trace_exception(5, "test_check_pattern [non-assertion]")

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Module for using Python abstract syntax tree (AST) support
# 
#

"""
Wrapper around AST. This is intended mainly for internal use, but a command-line example folows:

Sample usage:
   {script} __init__.py
"""

# Standard modules
import ast

# Installed modules
## TODO: import numpy

# Local modules
from mezcla import debug
from mezcla.main import Main
from mezcla import system
debug.trace(5, f"global __doc__: {__doc__}")
debug.assertion(__doc__)

# Constants for switches omitting leading dashes (e.g., DEBUG_MODE = "debug-mode")
TYPE_COMMENTS = "type-comments"

# Constants
TL = debug.TL

# Environment options
# Note: These are just intended for internal options, not for end users.
#
TODO_FUBAR = system.getenv_bool("TODO_FUBAR", False,
                                description="TODO:Fouled Up Beyond All Recognition processing")

class PythonAST:
    """Wrapper class around ast module"""

    def __init__(self, add_type_comments=None):
        """Initializer: optionally ADD_TYPE_COMMENTS"""
        debug.trace_fmtd(TL.VERBOSE, "PythonAST.__init__(): self={s}", s=self)
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")
        self.ast = None
        self.type_comments = add_type_comments

    def parse(self, python_code):
        """Parse PYTHON_CODE in AST"""
        self.ast = None
        try:
            self.ast = ast.parse(python_code, type_comments=self.type_comments)
        except:
            system.print_exception_info("parse")
        debug.trace_fmt(4, "parse({c}) => {p!r}", c=python_code, p=self.ast)
        return self.ast

    def dump(self):
        """Dump the AST into a string"""
        debug.assertion(self.ast)
        dump_text = ""
        try:
            dump_text = ast.dump(self.ast)
        except:
            system.print_exception_info("dump")
        debug.trace_fmt(4, "dump() => {d!r}", d=dump_text)
        return dump_text
        

class Script(Main):
    """Input processing class"""
    show_type_comments = False

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(TL.VERBOSE, "Script.setup(): self={s}", s=self)
        self.show_type_comments = self.get_parsed_option(TYPE_COMMENTS, self.show_type_comments)
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")

    def run_main_step(self):
        """Main processing step"""
        debug.trace_fmtd(5, "Script.run_main_step(): self={s}", s=self)
        input_python = self.read_entire_input()
        python_ast = PythonAST(add_type_comments=self.show_type_comments)
        python_ast.parse(input_python)
        print(python_ast.dump())


def main():
    """Entry point"""
    app = Script(
        description=__doc__.format(script=__file__),
        # note: reads entire file at once so manual input used (via run_main_step)
        skip_input=False, manual_input=True,
        boolean_options=[(TYPE_COMMENTS, "Include AST type comments")])
    app.run()
    
#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    main()



#!/usr/bin/env python3
#
# note: Tana's script for running Python tests as part of workflow
#


"""Master test script for shell-scripts repo"""

# Standard modules
import math
import subprocess
import os
from typing import Dict

# Installed modules
import yaml

# Local modules
from mezcla import debug
from mezcla.file_utils import read_yaml
from mezcla import glue_helpers as gh
from mezcla.glue_helpers import default_subtrace_level, TEMP_BASE, TEMP_FILE, PRESERVE_TEMP_FILE
from mezcla.my_regex import my_re
from mezcla import xml_utils
from mezcla import tpo_common as tpo
## TODO3: from mezcla import misc_utils
from mezcla import system
from  mezcla.system import to_int

# Environment options
# Note: These are just intended for internal options, not for end users.
#
THRESHOLDS_FILE = system.getenv_text(
    "THRESHOLDS_FILE",
    "thresholds.yaml",
    description="thresholds file for master_test.py",
)
TEST_REGEX = system.getenv_value(
    "TEST_REGEX", None,
    description="Regex for tests to include; ex: '^test_c.*' for debugging")
MYPY_TEST_REGEX = system.getenv_value(
    "MYPY_TEST_REGEX",
    TEST_REGEX,
    description="Regex for source files for mypy checks",
)
PYTEST_TEST_REGEX = system.getenv_value(
    "PYTEST_TEST_REGEX",
    TEST_REGEX,
    description="Regex for pytest",
)
MYPY_CONFIG_FILE = system.getenv_text(
    "MYPY_CONFIG_FILE", "mypy.ini",
    description="config file for mypy ",
)
MYPY_TEST_PATH = system.getenv_text(
    "MYPY_TEST_PATH",
    gh.dirname(__file__),
    description="directory with source files for mypy test",
)
PYTEST_TEST_PATH = system.getenv_text(
    "PYTEST_TEST_PATH", gh.form_path(gh.dirname(__file__), "tests"),
    description="directory with pytest tests"
)
MYPY_OUTPUT_PATH = system.getenv_text(
    "MYPY_OUTPUT_PATH", gh.form_path("output-files", "mypy_reports"),
    description="directory to save mypy reports in"
)
MYPY_WEIGHT = system.getenv_float(
    "MYPY_WEIGHT", 0.0,
    description="final weight for mypy tests"
)
MYPY_COMBINED = system.getenv_bool(
    "MYPY_COMBINED", False,
    description="Run mypy over all files at same time")
PYTEST_WEIGHT = system.getenv_float(
    "PYTEST_WEIGHT", (1.0 - MYPY_WEIGHT),
    description="final weight for pytest tests"
)
PYTEST_OPTIONS = system.getenv_value(
    "PYTEST_OPTIONS", None,
    description="Options for pytest"
)

VERBOSE_OUTPUT = system.getenv_bool(
    "VERBOSE_OUTPUT", False,
    description="Verbose output mode")

# -------------------------------------------------------------------------------
# Utility functions


def load_thresholds(filename):
    """Load test failure thresholds from a YAML file."""
    result = read_yaml(filename)
    debug.trace(5, f"load_thresholds({filename}) => {result!r}")
    return result


def round_p2str(num):
    """Round NUM using precision of 2"""
    # EX: round_p2str(1.678) => "1.68"
    # EX: round_p2str(1.6) => "1.60"
    return system.round_as_str(num, 2)


def run(command, trace_level=4, subtrace_level=None, **namespace) -> str :
    """Wrapper around subprocess.run
    Invokes COMMAND via system shell, using TRACE_LEVEL for debugging output, returning result. The command can use format-style templates, resolved from caller's namespace. The optional SUBTRACE_LEVEL sets tracing for invoked commands (default is same as TRACE_LEVEL); this works around problem with stderr not being separated, which can be a problem when tracing unit tests.
    Notes:
    - The result includes stderr, so direct if not desired (see issue):
          run("ls /tmp/fubar 2> /dev/null")
    - This is only intended for running simple commands. It would be better to create a subprocess for any complex interactions.
    - This function doesn't work fully under Win32. Tabs are not preserved, so redirect stdout to a file if needed.
    - If TEMP_FILE or TEMP_BASE defined, these are modified to be unique to avoid conflicts across processeses.
    - If OUTPUT, the result will be printed.
    """
    ## NOTE: streamlined version of gh.run
    debug.assertion(isinstance(trace_level, int))
    debug.trace(6, f"run({command}, tl={trace_level}, sub_tr={subtrace_level}")
    # Keep track of current debug level setting
    debug_level_env = os.getenv("DEBUG_LEVEL")
    if subtrace_level is None:
        subtrace_level = default_subtrace_level
    if subtrace_level != trace_level:
        system.setenv("DEBUG_LEVEL", str(subtrace_level))
    save_temp_base = TEMP_BASE
    if TEMP_BASE:
        system.setenv("TEMP_BASE", TEMP_BASE + "_subprocess_")
    save_temp_file = TEMP_FILE
    if TEMP_FILE and (PRESERVE_TEMP_FILE is not True):
        system.setenv("TEMP_FILE", TEMP_FILE + "_subprocess_")
    # Expand the command template
    # TODO: make this optional
    command_line = command
    if my_re.search("{.*}", command):
        command_line = tpo.format(command_line, indirect_caller=True, ignore_exception=False, **namespace)
    debug.trace(6 , f"issuing: {command_line}")
    # Run the command
    result = str(subprocess.run(command_line, check=False, shell=True))
    # Restore debug level setting in environment
    system.setenv("DEBUG_LEVEL", debug_level_env or "")
    system.setenv("TEMP_BASE", save_temp_base or "")
    if save_temp_file and (PRESERVE_TEMP_FILE is not True):
        system.setenv("TEMP_FILE", save_temp_file or "")
    debug.trace_fmt(7, "run(_) => {{\n{r}\n}}", r=gh.indent_lines(result))
    return result


def resolve_tools_file(filename):
    """Resolve full path for FILENAME, defaulting to <repo>/tools dir if not found"""
    full_path = filename
    if not system.file_exists(full_path):
        tools_dir = gh.form_path(gh.dirname(__file__), "..", "tools")
        debug.assertion(system.file_exists(gh.form_path(tools_dir, "run_tests.bash")))
        full_path = gh.form_path(tools_dir, filename)
    if not system.file_exists(full_path):
        full_path = gh.resolve_path(filename, heuristic=True)
    debug.trace(5, f"resolve_tools_file({filename}) => {full_path!r}")
    return full_path

# -------------------------------------------------------------------------------
# Main code

def run_mypy(thresholds: Dict[str, float]) -> int:
    """Run mypy and return the number of failures"""
    ## TODO1: rework so that failure is based on errors, not imprecision
    mypy_config_file = resolve_tools_file(MYPY_CONFIG_FILE)
    debug.assertion(system.file_exists(mypy_config_file))
    xml_report = None
    config = (
        f" --config-file {mypy_config_file}" if system.file_exists(mypy_config_file) else ""
    )
    common_mypy_options = "--check-untyped-defs --install-types --non-interactive"
    gh.full_mkdir(MYPY_OUTPUT_PATH)
    if MYPY_COMBINED:
        cmd = f"python -m mypy {MYPY_TEST_PATH} {config} --xml-report {MYPY_OUTPUT_PATH} {common_mypy_options}"
        run(cmd, shell=True, check=False)
    failed = 0

    # Read and parse xml report file
    report_file = gh.form_path(MYPY_OUTPUT_PATH, "index.xml")
    if MYPY_COMBINED and not system.file_exists(report_file):
        debug.trace(4, f"{report_file} not found, skipping mypy checks")
    else:
        if MYPY_COMBINED:
            xml_report = xml_utils.parse_xml(system.read_file(report_file))

        num_processed = 0
        for filename, threshold in thresholds.items():
            debug.trace_expr(4, filename, threshold)
            file_path = gh.resolve_path(filename)
            filename_proper = gh.basename(filename)

            # Exclude certain test files (e.g., files outside of the mezcla module)
            include = True
            if my_re.search(rf"^mezcla{os.sep}(.*__.*).+\.py$", filename):
                debug.trace(4, f"mypy: skipping dunder module {filename!r}")
                include = False
            elif MYPY_TEST_REGEX and not my_re.search(rf"{MYPY_TEST_REGEX}", file_path):
                debug.trace(5, f"Filtering test {file_path} not matching MYPY_TEST_REGEX ({MYPY_TEST_REGEX})",)
                include = False
            elif not system.file_exists(file_path):
                debug.trace(4, f"Warning: skipping missing module {filename!r}")
                include = False
            if not include:
                continue

            if VERBOSE_OUTPUT:
                if num_processed:
                    print("-" * 132)
                print(f"Running mypy over {filename_proper}")

            num_processed += 1
            if not MYPY_COMBINED:
                basename = gh.basename(filename_proper, ".py")
                mypy_output_path = gh.form_path(MYPY_OUTPUT_PATH, basename)
                cmd = f"python -m mypy {filename} {config} --xml-report {mypy_output_path} {common_mypy_options}"
                run(cmd, shell=True, check=False)
                report_file = gh.form_path(mypy_output_path, "index.xml")
                xml_report = xml_utils.parse_xml(system.read_file(report_file))
            
            # find current module name in report
            results = xml_report.find(f"./file[@name='{filename_proper}']")

            # calculate imprecise percentage and compare to threshold
            any_hints = to_int(results.get("any"))
            imprecise_hints = to_int(results.get("imprecise"))
            total_hints = to_int(results.get("total"))
            impreciseness = 0.0
            if total_hints > 0:
                impreciseness = ((any_hints + imprecise_hints) / total_hints) * 100
            debug.assertion(impreciseness <= threshold)
            debug.trace(
                5, f"{filename}, threshold: {threshold}, impreciseness: {impreciseness}"
            )
            if impreciseness > threshold:
                failed += 1

    return failed


def run_tests(thresholds: Dict[str, float]) -> int:
    """Run tests and compare the results with the allowed thresholds"""
    failed = 0
    for test_filename, threshold in thresholds.items():
        debug.trace_expr(4, test_filename, threshold)
        test_path = gh.resolve_path(test_filename)

        # Exclude certain test files (e.g., non-python test files listed in thresholds.yml)
        include = True
        if not my_re.search(r"^test_.*\.py$", gh.basename(test_filename)):
            debug.trace(5, f"Ignoring non-python test: {test_filename}")
            include = False
        elif PYTEST_TEST_REGEX and not my_re.search(rf"{PYTEST_TEST_REGEX}", test_filename):
            debug.trace(5, f"Filtering test {test_filename} not matching PYTEST_TEST_REGEX ({PYTEST_TEST_REGEX})",)
            include = False
        elif not system.file_exists(test_path):
            debug.trace(4, f"Warning: skipping missing test {test_path!r}")
            include = False
        if not include:
            continue

        # Collect test cases for the test
        ## BAD: cmd = f"pytest -k {test} --collect-only"
        cmd = f"pytest --collect-only {test_path}"
        collect_result = run(
            cmd, shell=True, text=True, capture_output=True, check=False
        )
        debug.trace_object(6, collect_result)
        total_tests = len(
            my_re.findall(
                r"<TestCaseFunction|<TestCaseClass|<Function|<Class",
                collect_result,
            )
        )
        # Compare against alternative way to detemine number of tests
        # ex: "=== 103 tests collected in 1.64s ==="
        if (total_tests == 0) or debug.debugging():
            summary_total_tests = 0
            if my_re.search(
                r"(\d+) tests collected", collect_result, flags=my_re.IGNORECASE
            ):
                summary_total_tests = system.to_int(my_re.group(1))
            debug.trace_expr(5, total_tests, summary_total_tests)
            debug.assertion(total_tests == summary_total_tests)
            total_tests = summary_total_tests if (total_tests == 0) else total_tests

        # Run tests for the test
        pytest_options = (PYTEST_OPTIONS or "")
        cmd = f"pytest {pytest_options} {test_path}"
        run_result = run(cmd)
        debug.trace_object(6, run_result)
        failed_tests = len(my_re.findall(r"FAILED", run_result))
        debug.assertion(failed_tests <= total_tests)
        # Compare against alternative way to detemine number of failures
        # ex: "=== 12 passed, 49 skipped, 13 xfailed, 29 xpassed in 4.76s ==="
        if debug.debugging():
            summary_failed_tests = 0
            if my_re.search(r"(\d+) failed", run_result, flags=my_re.IGNORECASE):
                summary_failed_tests = system.to_int(my_re.group(1))
            debug.trace_expr(5, failed_tests, summary_failed_tests)
            debug.assertion(failed_tests == summary_failed_tests)

        # Calculate the number of allowed failures
        # note: threshold is for success (e.g., 51 means 51% of tests passed)
        # TODO3: have sanity checks account for minor floatiing point differences
        required_successes = (
            math.ceil(total_tests * threshold / 100) if threshold else 0
        )
        debug.assertion(0 <= required_successes <= total_tests)
        allowed_failures = max(1 - required_successes, 0)

        # Check if the number of failed tests exceeds the allowed threshold
        if total_tests == 0:
            print(f"Warning: No tests were found for test {test_path}.")
            debug.trace_expr(5, collect_result, collect_result)
            continue
        module_failure = failed_tests and (failed_tests <= allowed_failures)
        failed_percent = round_p2str(failed_tests / total_tests * 100)
        success_percent = round_p2str(100.0 - (failed_tests / total_tests * 100))
        ## TODO3: debug.assertion(misc_utils.is_close(system.to_float(failed_percent) + system.to_float(success_percent), 100))
        debug.trace_expr(
            6,
            failed_tests,
            allowed_failures,
            total_tests,
            module_failure,
            required_successes,
        )

        # Format message to stdout: either error, warning or FYI on test summary.
        # note: format shows success rate to match batspp_report.py
        label = "Error" if module_failure else "Warning" if failed_tests else "FYI"
        ## OLD:
        ## print(f"{label}: {test_filename} {failed_tests} of {total_tests} tests failed ({failed_percent}%)",
        ##      end="")
        print(
            f"{label}: {test_filename} {failed_tests} of {total_tests} tests failed: {failed_percent}%; success ({success_percent}%); threshold={threshold}%",
            end="",
        )
        if module_failure:
            num_good = total_tests - failed_tests
            short = required_successes - num_good
            debug.assertion(0 <= required_successes <= total_tests)
            debug.assertion(0 <= short <= (total_tests - failed_tests) <= total_tests)
            print(
                (
                    f": {short} short of the {required_successes} required successes"
                    + f" (i.e., {round_p2str(threshold)}+%)"
                ),
                end="",
            )
            failed += 1
        print(".")
    return failed

    # Return amount of failed modules


# -------------------------------------------------------------------------------


def main():
    """Main function"""

    # Load the thresholds from the YAML file, falling back to defaults for all test files
    # in the tests directory (TODO: merge the two sources with file trumping default).
    # note: uses default of 25% succeses allowed just for sake of getting tests operational
    # under Github actions (TODO: lower to 50%).
    thresholds_path = resolve_tools_file(THRESHOLDS_FILE)
    mypy_thresholds = {
        module: 40.0 for module in gh.get_matching_files(f"{MYPY_TEST_PATH}/*.py")
    }
    test_thresholds = {
        test_file: 25.0 for test_file in gh.get_matching_files(f"{PYTEST_TEST_PATH}/*.py")
    }
    # note: thresholds = mypy_thresholds | test_thresholds for python 3.9+
    default_thresholds = {**mypy_thresholds, **test_thresholds}
    debug.trace_expr(5, default_thresholds, prefix="default thresholds: ")
    if system.file_exists(thresholds_path):
        new_thresholds: Dict[str, float] = load_thresholds(thresholds_path)
        for filename, threshold in new_thresholds.items():
            if my_re.search(r"^.*(tests{os.sep}).*", filename):
                test_thresholds[filename] = threshold
            elif my_re.search(r"^.*\.py$", filename):
                mypy_thresholds[filename] = threshold
            else:
                debug.trace(4, f"Warning: ignoring threshold for file {filename!r}")
    else:
        debug.trace(2, f"Warning: unable to find {thresholds_path}")
    debug.trace_expr(
        5, {**mypy_thresholds, **test_thresholds}, prefix="final thresholds"
    )

    # Run tests and compare the results with the allowed thresholds
    test_failures = 0
    mypy_failures = 0    
    if PYTEST_WEIGHT > 0:
        test_failures = run_tests(test_thresholds)
    if MYPY_WEIGHT > 0:
        mypy_failures = run_mypy(mypy_thresholds)
    failed = (mypy_failures * MYPY_WEIGHT) + (test_failures * PYTEST_WEIGHT)

    message = "All OK"
    code = 0
    if failed > 0:
        code = failed
        message = f"Error: {test_failures} module failed, and {mypy_failures} mypy problems: combined score {failed:.2f}"
    system.exit(message, status_code=code)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
#
# Test(s) for ../__init__.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test___init__.py
#

"""Tests for __init__ module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import debug
from mezcla.my_regex import my_re

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                  global module object
#    TestTemplate.script_module:  path to file
import mezcla.__init__ as THE_MODULE

class TestIt(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_data_file(self):
        """Makes sure TODO works as expected"""
        debug.trace(4, f"TestIt.test_data_file(); self={self}")
        output = self.run_script("", self.temp_file)
        assert(not output.strip())
        return

class TestIt2:
    """Another class for testcase definition
    Note: Needed to avoid error with pytest due to inheritance with unittest.TestCase via TestWrapper"""

    def test_version(self):
        """Test version string"""
        debug.trace(5, f"test_version(); self={self}")
        assert(hasattr(THE_MODULE, "__VERSION__"))
        assert(my_re.match(r"^\d+.\d+\.\d+.*", THE_MODULE.__VERSION__))


#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# sys_version_info_hack.py: hack for redefining sys.version_info to be python3 compatible
#
# python3:
# >>> sys.version_info
# sys.version_info(major=3, minor=6, micro=5, releaselevel='final', serial=0)
# >>> type(sys.version_info)
# <class 'sys.version_info'>
#
# python2 (pre 2.7):
# In [11]: sys.version_info
# Out[11]: (2, 6, 7, 'final', 0)
# In [12]: type(sys.version_info)
# Out[12]: tuple
#
# usage:
#    import sys_version_info_hack
#

"""Redefines sys.version_info to be python3 compatible"""

# Standard packages
import os
import re
import sys

# Installed packages
# n/a

# Local packages
# None (since most rely upon debug.py)
## OLD: import debug


if sys.version_info[0] < 3:
    
    SYS_VERSION_INFO_OLD = sys.version_info
    
    class sys_version_info(object):
        """Class to mimic Python3 version_info"""
    
        def __init__(self, major=SYS_VERSION_INFO_OLD[0],
                     minor=SYS_VERSION_INFO_OLD[1],
                     micro=SYS_VERSION_INFO_OLD[2],
                     releaselevel=SYS_VERSION_INFO_OLD[3],
                     serial=SYS_VERSION_INFO_OLD[4]):
            """Constructor"""
            self.sys_version_info_old = SYS_VERSION_INFO_OLD
            self.major = major
            self.minor = minor
            self.micro = micro
            self.releaselevel = releaselevel
            self.serial = serial
    
        def __getitem__(self, index):
            """Return old-style INDEX into sys.version_info"""
            return  self.sys_version_info_old[index]
    
    # Replace old-style sys.version_info with new-style
    ## OLD: if isinstance(sys.version_info, tuple):
    if type(sys.version_info) == tuple:   # pylint: disable=unidiomatic-typecheck
        ## OLD: debug.trace(4, "Warning: replacing sys version_info w/ python3-like version")
        if re.match(r"[1-9]+", os.environ.get("DEBUG_LEVEL")):
            sys.stderr.write("Warning: replacing sys version_info w/ python3-like version\n")
        sys.version_info = sys_version_info()

#-------------------------------------------------------------------------------

## OLD: def main(args):
def main(_args):
    """Supporting code for command-line processing"""
    ## OLD: debug.trace(2, "Warning: not intended for command-line use")
    sys.stderr.write("Warning: not intended for command-line use\n")

if __name__ == '__main__':
    main(sys.argv)



#! /usr/bin/env python3
#
# Based on following:
#   https://stackoverflow.com/questions/552744/how-do-i-profile-memory-usage-in-python
#
# TODO:
# - Add muiltithreading support.
#

"""Illustration of tracemalloc module (for reading word list)"""

# Standard modules
from collections import Counter
import linecache
import os
import tracemalloc

# Local modules
from mezcla import debug
from mezcla.main import Main
from mezcla import system

NUM_ITEMS = system.getenv_int("NUM_ITEMS", 10,
                              "Number ot items to display")
PREFIX_LEN = system.getenv_int("PREFIX_LEN", 3,
                               "Prefix length for item display")
WORDLIST_FILE = system.getenv_text("WORDLIST_FILE", "/usr/share/dict/american-english",
                                   "File with word list")

class TraceMalloc:
    """Wrapper class around tracemalloc"""

    def __init__(self, limit=None):
        """Class intiializer"""
        debug.trace(4, "TraceMalloc.__init__()")
        self.snapshot = None
        self.limit = limit
        tracemalloc.start()
    
    
    def display_top(self, key_type=None, limit=None):
        """Show top usages"""
        debug.trace(4, f"TraceMallocdisplay_top([{key_type}, {limit}])")
        if key_type is None:
            key_type = 'lineno'
        if limit is None:
            limit = (self.limit or NUM_ITEMS)
        snapshot = self.snapshot.filter_traces((
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<unknown>"),
        ))
        top_stats = snapshot.statistics(key_type)
    
        print("Top %s memory allocations" % limit)
        for index, stat in enumerate(top_stats[:limit], 1):
            frame = stat.traceback[0]
            # replace "/path/to/module/file.py" with "module/file.py"
            filename = os.sep.join(frame.filename.split(os.sep)[-2:])
            print("#%s: %s:%s: %.1f KiB"
                  % (index, filename, frame.lineno, stat.size / 1024))
            line = linecache.getline(frame.filename, frame.lineno).strip()
            if line:
                print('    %s' % line)
    
        other = top_stats[limit:]
        if other:
            size = sum(stat.size for stat in other)
            print("%s other: %.1f KiB" % (len(other), size / 1024))
        total = sum(stat.size for stat in top_stats)
        print("Total allocated size: %.1f KiB" % (total / 1024))

    def take_snapshot(self):
        """Take snapshot or memory usage"""
        self.snapshot = tracemalloc.take_snapshot()
    
    def display(self, limit=None):
        """Display current snapshot"""
        debug.trace(3, "TraceMalloc.display()")

        if not self.snapshot:
            self.take_snapshot()
        self.display_top(limit=limit)

#-------------------------------------------------------------------------------
        
def main():
    """Entry point for script"""
    debug.trace(4, "main()")

    # Show simple usage if --help given
    dummy_app = Main(description=__doc__, skip_input=False, manual_input=False)
    debug.assertion(dummy_app.parsed_args)

    trace_malloc = TraceMalloc(limit=NUM_ITEMS)
    
    counts = Counter()
    fname = WORDLIST_FILE
    with system.open_file(fname) as wordlist_file:
        words = list(wordlist_file)
        for word in words:
            prefix = word[:PREFIX_LEN]
            counts[prefix] += 1
    print('Top prefixes:', counts.most_common(3))

    trace_malloc.display()
    
#-------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#! /usr/bin/env python3
#
# Test(s) for ../<module>.py
#
# Notes:
# - This can be run as follows (e.g., from root of repo):
#   $ pytest ./mezcla/tests/test_<module>.py
#

"""Tests for introspection module"""

# Standard modules
## TODO: from collections import defaultdict

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
THE_MODULE = None
try:
    import mezcla.introspection as THE_MODULE
except:
    system.print_exception_info("introspection import") 

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    # note: script_module used in argument parsing sanity check (e.g., --help)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_01_simple_introspection(self):
        """Test for simple introspection"""
        debug.trace(4, f"TestIt.test_01_simple_introspection(); self={self}")
        fubar = 123.321
        fubar_expr = THE_MODULE.intro.format(fubar)
        self.do_assert(my_re.search("fubar.*123.321", fubar_expr))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_02_multiline_introspection(self):
        """Test for multi-line introspection"""
        debug.trace(4, f"TestIt.test_02_multiline_introspection(); self={self}")
        multiline_value_expr = THE_MODULE.intro.format(
            2
            +
            2
            ==
            5)
        debug.trace_expr(5, multiline_value_expr)
        self.do_assert(my_re.search(r"2.*\+.*2.*==.*5", multiline_value_expr,
                                    flags=my_re.DOTALL|my_re.MULTILINE))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_03_prefix(self):
        """Test for prefix added to introspection result"""
        debug.trace(4, f"TestIt.test_03_prefix(); self={self}")
        var = 456
        var_expr = THE_MODULE.intro.format(var, _prefix="here: ")
        assert my_re.search("here: var.*456", var_expr)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_04_max_len(self):
        """Test for introspection truncation"""
        debug.trace(4, f"TestIt.test_04_max_len(); self={self}")
        var = "-" * 123
        var_expr = THE_MODULE.intro.format(var, max_len=4)
        ## TODO2: assert my_re.search("var='----\.\.\.'", var_expr)
        ##                                            ^
        assert my_re.search("var='----\.\.\.", var_expr)
        return


#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
# 
# Evaluate '#: EX' tests placed in python source. This reformats EX comments
# into the doctest text format. For example,
#    # EX: is_symbolic("3.14159") => False
# gets converted as follows:
#    >>> is_symbolic("3.14159")
#    False
#

"""
Convert example comments in Python scource to doctest format and evaluate.

Sample usage:
   {script} text_utils.py
"""

# Standard modules
from types import ModuleType            # pylint: disable=unused-import

# Installed modules
## TODO2: import ast_tools

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla.my_regex import my_re
from mezcla import system

# Constants for switches, omitting leading dashes
OUTPUT_ARG = "output"                   # TODO: => just-output

# Constants
TL = debug.TL

# Environment options
SKIP_BACKGROUND = system.getenv_bool(
    "SKIP_BACKGROUND", False,
    description="Skip example tests involving background jobs")
NORMALIZE_RESULT = system.getenv_bool(
    "NORMALIZE_RESULT", False,
    description="Normalize result to acount for doctest quirks as with quotes")


class TestConverter:
    """Class to convert example EX-style comments into doctest text format"""

    def __init__(self):
        """Initializer"""
        debug.trace_fmtd(TL.VERBOSE, "Helper.__init__(): self={s}", s=self)
        self.test_text = ""
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")

    def get_tests(self):
        """Returns set of tests as string"""
        result = self.test_text
        debug.trace(6, f"get_tests() => {result}")
        return result

    def add_module(self, module_spec):
        """Add import-* spec from MODULE_SPEC"""
        # pylint: disable=eval-used, exec-used
        ## TODO2: debug.assertion(eval("isintance(eval({module_spec!r}, module)"))
        exec(f"import {module_spec}")
        debug.assertion(eval(f"isinstance({module_spec}, ModuleType)"))
        self.test_text += f">>> from {module_spec} import *\n\n"

    def convert(self, line, line_num):
        """Convert example test in LINE at LINE_NUM"""
        expression = result = None
        OK = False

        # Parse test expression
        if my_re.search(r"EX: (.*)\s*=>\s*(.*)", line):
            expression, result = my_re.groups()
            OK = True
        elif my_re.search(r"EX: (.*)", line):
            expression = my_re.group(1)
            result = "True"
            OK = True
        else:
            pass

        # Exclude certain commands
        if expression and SKIP_BACKGROUND and my_re.search(r"run|issue(.*&)", expression):
            debug.trace(4, f"Ignoring invocation of background job at {line_num}: {line}")
            OK = False

        # Normalize result
        ## TODO3: my_re.search(r'^".*[^"].*"$', result)
        if result and NORMALIZE_RESULT and my_re.search(r'^".*"$', result):
            result = f'{result[1:-1]}'

        # Add to tests
        if ((not OK) or ((expression is None) and (result is None))):
            debug.trace(6, f"Ignoring line {line_num}: {line}")
        else:
            self.test_text += f">>> {expression.strip()}\n{result.strip()}\n\n"
        return OK


class Script(Main):
    """Input processing class"""
    output_arg = False
    exc = None
    num_tests = 0

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(TL.VERBOSE, "Script.setup(): self={s}", s=self)
        self.output_arg = self.get_parsed_option(OUTPUT_ARG, self.output_arg)
        self.exc = TestConverter()
        module = gh.basename(self.filename, ".py")
        self.exc.add_module(module)
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")

    def process_line(self, line):
        """Processes current line looking for EX comments and converting"""
        debug.trace_fmtd(TL.QUITE_DETAILED, "Script.process_line({l})", l=line)
        ok = self.exc.convert(line, self.line_num)
        if ok:
            self.num_tests += 1

    def wrap_up(self):
        """Output the final set of tests"""
        debug.trace_fmtd(TL.VERBOSE, "Script.wrap_up(): self={s}", s=self)
        test_spec = self.exc.get_tests()
        if self.output_arg:
            print(test_spec)
        elif self.num_tests:
            system.write_file(self.temp_file, test_spec)
            print(gh.run(f"python -m doctest -v {self.temp_file}"))
        else:
            debug.trace(4, "FYI: No tests extracted")
        
def main():
    """Entry point"""
    app = Script(
        description=__doc__.format(script=gh.basename(__file__)),
        skip_input=False, manual_input=False,
        boolean_options=[(OUTPUT_ARG, "Output converted tests without running")])
    app.run()
    
#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    debug.trace(5, f"module __doc__: {__doc__}")
    main()
